# -*- coding: utf-8 -*-
""" A module with classes for aggregation.
An Aggregator has two methods: process and finish.

process(group, val) is called to "add" val to the aggregation of the set of 
values identified by the value of group. The value in group (which could be any 
hashable type, also a tuple as ('A', 'B')) thus corresponds to the GROUP BY 
attributes in SQL.

finish(group, default) is called to get the final result for group. If no such
results exists, default is returned.
"""

__author__ = "Christian Thomsen"
__maintainer__ = "Sílex Sistemas Ltda."
__version__ = '0.0.1'
__all__ = ['Aggregator', 'SimpleAggregator', 'Sum', 'Count', 'CountDistinct',
           'Max', 'Min', 'Avg']


class Aggregator(object):
    def process(self, group, val):
        raise NotImplementedError

    def finish(self, group, default=None):
        raise NotImplementedError


class SimpleAggregator(Aggregator):
    def __init__(self):
        self._results = {}

    def process(self, group, val):
        pass

    def finish(self, group, default=None):
        return self._results.get(group, default)


class Sum(SimpleAggregator):
    def process(self, group, val):
        tmp = self._results.get(group, 0)
        tmp += val
        self._results[group] = tmp


class Count(SimpleAggregator):
    def process(self, group, val):
        tmp = self._results.get(group, 0)
        tmp += 1
        self._results[group] = tmp


class CountDistinct(SimpleAggregator):
    def process(self, group, val):
        if group not in self._results:
            self._results[group] = set()
        self._results[group].add(val)

    def finish(self, group, default=None):
        if group not in self._results:
            return default
        return len(self._results[group])
    

class Max(SimpleAggregator):
    def process(self, group, val):
        if group not in self._results:
            self._results[group] = val
        else:
            tmp = self._results[group]
            if val > tmp:
                self._results[group] = val


class Min(SimpleAggregator):
    def process(self, group, val):
        if group not in self._results:
            self._results[group] = val
        else:
            tmp = self._results[group]
            if val < tmp:
                self._results[group] = val


class Avg(Aggregator):
    def __init__(self):
        self.__sum = Sum()
        self.__count = Count()

    def process(self, group, val):
        self.__sum.process(group, val)
        self.__count.process(group, val)

    def finish(self, group, default=None):
        tmp = self.__sum.finish(group, None)
        if tmp is None:
            return default
        else:
            return float(tmp) / self.__count(group)



# -*- coding: utf-8 -*-
"""This module contains methods and classes for making parallel ETL flows.
   Warning: This is still experimental and things may be changed drastically.
   If you have ideas, comments, bug reports, etc., please report them to
   Christian Thomsen (chr@cs.aau.dk)
   Note that this module in many cases will give better results with Jython
   (where it uses threads) than with CPython (where it uses processes).
"""
__author__ = "Christian Thomsen"
__maintainer__ = "Sílex Sistemas Ltda."
__version__ = '0.0.1'
__all__ = ['splitpoint', 'endsplits', 'createflow', 'Decoupled', \
               'shareconnectionwrapper', 'getsharedsequencefactory']

import copy
import os
from Queue import Empty
import sys
if sys.platform.startswith('java'):
    # Jython specific code in jythonmultiprocessing
    import pygrametl.jythonmultiprocessing as multiprocessing
else:
    # Use (C)Python's std. lib.
    import multiprocessing

import pygrametl


# Support for spawned processes to be able to terminate all related processes
# in case of an uncaught exception

_masterpid = os.getpid() # the first to import parallel
_toterminator = None
def _getexitfunction():
    """Return a function that halts the execution of pygrametl.

       pygrametl uses the function as excepthook in spawned processes such that
       an uncaught exception halts the entire execution.
    """
    # On Java, System.exit will do as there are no separate processes
    if sys.platform.startswith('java'):
        def javaexitfunction():
            import java.lang.System
            java.lang.System.exit(1)
        return javaexitfunction
    
    # else see if the os module provides functions to kill process groups;
    # this should be the case on UNIX.
    import signal
    if hasattr(os, 'getpgrp') and hasattr(os, 'killpg'):
        def unixexitfunction():
            procgrp = os.getpgrp()
            os.killpg(procgrp, signal.SIGTERM)
        return unixexitfunction

    # else, we are on a platform that does not allow us to kill a group.
    # We make a special process that gets the pids of all calls to
    # this procedure. The function we return, informs this process to kill
    # all processes it knows about.

    # set up the terminator
    global _toterminator
    if _toterminator is None:
        _toterminator = multiprocessing.Queue()
        def terminatorfunction():
            pids = set([_masterpid])
            while True:
                item = _toterminator.get()
                if type(item) == int:
                    pids.add(item)
                else:
                    # We take it as a signal to kill all
                    for p in pids:
                        os.kill(p, 9) # we don't know which signals exist; use 9
                    return

        terminatorprocess = multiprocessing.Process(target=terminatorfunction)
        terminatorprocess.daemon = True
        terminatorprocess.start()

    # tell the terminator about this process
    _toterminator.put(os.getpid())

    # return a function that tells the terminator to kill all known processes
    def exitfunction():
        _toterminator.put('TERMINATE')

    return exitfunction


def _getexcepthook():
    "Return a function that can be used as except hook for uncaught exceptions."
    if not sys.argv[0]:
        # We are in interactive mode and don't want to terminate
        return sys.excepthook
    # else create a function that terminates all spawned processes and this
    # in case of an uncaught exception
    exit = _getexitfunction()
    def excepthook(exctype, excvalue, exctraceback):
        import traceback
        sys.stderr.write(
            "An uncaught exception occured. Terminating pygrametl.\n")
        traceback.print_exception(exctype, excvalue, exctraceback)
        exit()
    return excepthook


# Stuff for @splitpoint 

splitno = None
def _splitprocess(func, input, output, splitid):
    # The target of a process created for a splitpoint
    global splitno
    splitno = splitid
    sys.excepthook = _getexcepthook() # To handle uncaught exceptions and halt
    (args, kw) = input.get()
    while True:
        res = func(*args, **kw)
        if output is not None:
            output.put(res)
        input.task_done()
        (args, kw) = input.get()

_splitpointqueues = []
def splitpoint(*arg, **kwargs):
    """To be used as an annotation to make a function run in a separate process.

       Each call of a @splitpoint annotated function f involves adding the 
       request (and arguments, if any) to a shared queue. This can be 
       relatively expensive if f only uses little computation time. 
       The benefits from @splitpoint are thus best obtained for a function f 
       which is time-consuming. To wait for all splitpoints to finish their
       computations, call endsplits().

       @splitpoint can be used as in the following examples:

       @splitpoint
       def f(args):
          # The simplest case. Makes f run in a separate process.
          # All calls of f will return None immediately and f will be
          # invoked in the separate process.
          ...

       @splitpoint()
       def g(args):
          # With parentheses. Has the same effect as the previous example.
          ...

       @splitpoint(output=queue, instances=2, queuesize=200)
       def h(args):
          # With keyword arguments. It is not required that
          # all of keyword arguments above are given.
          ...
          
       Keyword arguments:
       - output: If given, it should be a queue-like object (offering the
         .put(obj) method). The annotated function's results will then be put
         in the output
       - instances: Determines how many processes should run the function.
         Each of the processes will have the value parallel.splitno set to
         a unique value between 0 (incl.) and instances (excl.).
       - queuesize: Given as an argument to a multiprocessing.JoinableQueue
         which holds arguments to the annotated function while they wait for
         an idle process that will pass them on to the annotated function.
         The argument decides the maximum number of calls that can wait in the
         queue. 0 means unlimited. Default: 0
    """

    # We construct a function called decorator. We either return this
    # decorator or the decorator applied to the function to decorate.
    # It depends on the user's arguments what to do:
    #
    # When the user uses parentheses after the annotation (as in
    # "@splitpoint(output=x)" or even "@splitpoint()"), arg automatically
    # becomes empty, i.e., arg == (). In that case we return the created
    # decorator such that Python can use it to decorate some function (which
    # we don't know.
    #
    # When no arguments are given (as in "@splitpoint"), arg has a 
    # single element, namely the function to annotate, arg == (<function>,).
    # We then return decorator(function).

    for kw in kwargs.keys():
        if kw not in ('instances', 'output', 'queuesize'):
            raise TypeError, \
                "'%s' is an invalid keyword argument for splitpoint" % kw

    output = kwargs.get('output', None)
    instances = kwargs.get('instances', 1)
    queuesize = kwargs.get('queuesize', 0)

    def decorator(func):
        global _splitpointqueues
        if instances < 1:
            # A special case where there is no process so
            # we just call func directly
            def sillywrapper(*args, **kw):
                res = func(*args, **kw)
                if output is not None:
                    output.put(res)
            return sillywrapper
        # Else set up processes
        input = multiprocessing.JoinableQueue(queuesize)
        for n in range(instances):
            p = multiprocessing.Process(target=_splitprocess,\
                                            args=(func, input, output, n))
            p.name = 'Process-%d for %s' % (n, func.__name__)
            p.daemon = True
            p.start()
        _splitpointqueues.append(input)
        def wrapper(*args,**kw):
            input.put((args, kw))
        return wrapper

    if len(arg) == 0:
        return decorator
    elif len(arg) == 1:
        return decorator(*arg)
    else:
        raise ValueError, 'More than one *arg given'


def endsplits():
    """Wait for all splitpoints to finish"""
    global _splitpointqueues
    for q in _splitpointqueues:
        q.join()


# Stuff for (function) flows

def _flowprocess(func, input, output, inclosed, outclosed):
    sys.excepthook = _getexcepthook() # To handle uncaught exceptions and halt
    retryingafterclose = False
    while True:
        try:
            batch = input.get(True, 0.1)
            for args in batch:
                func(*args)
            input.task_done()
            output.put(batch)
        except Empty:
            if not inclosed.value:
                # A new item may be on its way, so try again
                continue
            elif not retryingafterclose:
                # After the get operation timed out, but before we got to here,
                # an item may have been added before the closed mark got set,
                # so we have to try again
                retryingafterclose = True
                continue
            else:
                # We have now tried get again after we saw the closed mark.
                # There is no more data.
                break
    output.close()
    outclosed.value = 1


class Flow(object):
    """A Flow consists of different functions running in different processes.
       A Flow should be created by calling createflow. 
    """
    def __init__(self, queues, closedarray, batchsize=1000):
        self.__queues = queues
        self.__closed = closedarray
        self.__batchsize = batchsize
        self.__batch = []
        self.__resultbatch = []

    def __iter__(self):
        try:
            while True:
                yield self.get()
        except Empty:
            return

    def __call__(self, *args):
        self.process(*args)
        
    def process(self, *args):
        "Insert arguments into the flow"
        self.__batch.append(args)
        if len(self.__batch) == self.__batchsize:
            self.__queues[0].put(self.__batch)
            self.__batch = []

    def __oneortuple(self, thetuple):
        # If there is only one element in the given tuple, return that element;
        # otherwise return the full tuple.
        if len(thetuple) == 1:
            return thetuple[0]
        return thetuple

    def get(self):
        """Return the result of a single call of the flow.

           If the flow was called with a single argument -- as in 
           flow({'foo':0, 'bar':1}) -- that single argument is returned (with
           the side-effects of the flow preserved).

           If the flow was called with multiple arguments -- as in
           flow({'foo'0}, {'bar':1}) -- a tuple with those arguments is
           returned (with the side-effects of the flow preserved).
        """
        if self.__resultbatch:
            return self.__oneortuple(self.__resultbatch.pop())
        
        # Else fetch new data from the queue
        retryingafterclose = False
        while True:
            try:
                tmp = self.__queues[-1].get(True, 0.1)
                self.__queues[-1].task_done()
                tmp.reverse()
                self.__resultbatch = tmp
                return self.__oneortuple(self.__resultbatch.pop())
            except Empty:
                # See explanation in _flowprocess
                if not self.__closed[-1].value:
                    continue
                elif not retryingafterclose:
                    retryingafterclose = True
                    continue
                else:
                    raise Empty

    def getall(self):
        """Return all results in a single list. 

        The results are of the same form as those returned by get.
        """
        res = []
        try:
            while True:
                res.append(self.get())
        except Empty:
            pass
        return res

    def join(self):
        "Wait for all queues to be empty, i.e., for all computations to be done"
        for q in self.__queues:
            q.join()

    def close(self):
        "Close the flow. New entries can't be added, but computations continue."
        if self.__batch:
            self.__queues[0].put(self.__batch)
            self.__batch = [] # Not really necessary, but ...
        self.__queues[0].close()
        self.__closed[0].value = 1

    @property
    def finished(self):
        "Tells if the flow is closed and all computations have finished"
        for v in self.__closed:
            if not v.value:
                return False
        return True


def _buildgroupfunction(funcseq):
    def groupfunc(*args):
        for f in funcseq:
            f(*args)
        groupfunc.__doc__ = 'group function calling ' + \
            (', '.join([f.__name__ for f in funcseq]))
    return groupfunc

        
def createflow(*functions, **options):
    """Create a flow of functions running in different processes.

       A Flow object ready for use is returned.

       A flow consists of several functions running in several processes.
       A flow created by 
           flow = createflow(f1, f2, f3) 
       uses three processes. Data can be inserted into the flow by calling it
       as in flow(data). The argument data is then first processed by f1(data),
       then f2(data), and finally f3(data). Return values from f1, f2, and f3 
       are *not* preserved, but their side-effects are. The functions in a flow
       should all accept the same number of arguments (*args are also okay).

       Internally, a Flow object groups calls together in batches to reduce
       communication costs (see also the description of arguments below).
       In the example above, f1 could thus work on one batch, while f2 works
       on another batch and so on. Flows are thus good to use even if there
       are many calls of relatively fast functions.

       When no more data is to be inserted into a flow, it should be closed
       by calling its close method.

       Data processed by a flow can be fetched by calling get/getall or simply
       iterating the flow. This can both be done by the process that inserted
       data into the flow or by another (possibly concurrent) process. All
       data in a flow should be fetched again as it otherwise will remain in 
       memory .

       Arguments:
       - *functions: A sequence of functions of sequences of functions.
         Each element in the sequence will be executed in a separate process.
         For example, the argument (f1, (f2, f3), f4) leads to that
         f1 executes in process-1, f2 and f3 execute in process-2, and f4
         executes in process-3.
         The functions in the sequence should all accept the same number of
         arguments.
       - **options: keyword arguments configuring details. The considered
         options are:
         - batchsize: an integer deciding how many function calls are "grouped
           together" before they are passed on between processes. The default
           is 500.
         - queuesize: an integer deciding the maximum number of batches
           that can wait in a JoinableQueue between two different processes. 
           0 means that there is no limit.
           The default is 25.
       
    """
    # A special case
    if not functions:
        return Flow([multiprocessing.JoinableQueue()],\
                        [multiprocessing.Value('b', 0)], 1)

    # Create functions that invoke a group of functions if needed
    resultfuncs = []
    for item in functions:
        if callable(item):
            resultfuncs.append(item)
        else:
            # Check the arguments
            if not hasattr(item, '__iter__'):
                raise ValueError, \
                    'An element is neither iterable nor callable'
            for f in item:
                if not callable(f):
                    raise ValueError, \
                        'An element in a sequence is not callable'
            # We can - finally - create the function
            groupfunc = _buildgroupfunction(item)
            resultfuncs.append(groupfunc)

    # resultfuncs are now the functions we need to deal with.
    # Each function in resultfuncs should run in a separate process
    queuesize = ('queuesize' in options and options['queuesize']) or 0
    batchsize = ('batchsize' in options and options['batchsize']) or 25
    if batchsize < 1:
        batchsize = 25
    queues = [multiprocessing.JoinableQueue(queuesize) for f in resultfuncs]
    queues.append(multiprocessing.JoinableQueue(queuesize)) # for the results
    closed = [multiprocessing.Value('b', 0) for q in queues] # in shared mem
    for i in range(len(resultfuncs)):
        p = multiprocessing.Process(target=_flowprocess, \
                                        args=(resultfuncs[i], \
                                                  queues[i], queues[i+1], \
                                                  closed[i], closed[i+1]))
        p.start()
        
    # Now create and return the object which allows data to enter the flow
    return Flow(queues, closed, batchsize)


### Stuff for Decoupled objects

class FutureResult(object):
    """Represent a value that may or may not be computed yet. 
    FutureResults are created by Decoupled objects.
    """

    def __init__(self, creator, id):
        """Arguments:
        - creator: a value that identifies the creator of the FutureResult.
          Use a primitive value.
        - id: a unique identifier for the FutureResult.
        """
        self.__creator = creator
        self.__id = id

    @property
    def creator(self):
        return self.__creator

    @property
    def id(self):
        return self.__id

    def __setstate__(self, state):
        self.__creator = state[0]
        self.__id = state[1]

    def __getstate__(self):
        return (self.__creator, self.__id)


# TODO: Add more documentation for developers. Users should use Decoupled 
# through its subclasses DecoupledDimension and DecoupledFactTable in 
# pygrametl.tables

class Decoupled(object):
    __instances = []

    def __init__(self, obj, returnvalues=True, consumes=(),
                 directupdatepositions=(),
                 batchsize=500, queuesize=200, autowrap=True):
        self.__instancenumber = len(Decoupled.__instances)
        self.__futurecnt = 0
        Decoupled.__instances.append(self)
        self._obj = obj
        if hasattr(obj, '_decoupling') and callable(obj._decoupling):
            obj._decoupling()
        self.batchsize = batchsize
        self.__batch = []
        self.__results = {}
        self.autowrap = autowrap
        self.__toworker = multiprocessing.JoinableQueue(queuesize)
        if returnvalues:
            self.__fromworker = multiprocessing.JoinableQueue(queuesize)
        else:
            self.__fromworker = None
        self.__otherqueues = dict([(dcpld.__instancenumber, dcpld.__fromworker)\
                                       for dcpld in consumes])
        self.__otherresults = {} # Will store dicts - see also __decoupledworker
        self.__directupdates = directupdatepositions
        
        self.__worker = multiprocessing.Process(target=self.__decoupledworker)
        self.__worker.daemon = True
        self.__worker.name = 'Process for %s object for %s' % \
            (self.__class__.__name__, getattr(obj, 'name', 'an unnamed object'))
        self.__worker.start()


    ### Stuff for the forked process
    def __getresultfromother(self, queuenumber, id):
        while True:
            if id in self.__otherresults[queuenumber]:
                return self.__otherresults[queuenumber].pop(id)
            # else wait for more results to become available
            self.__otherresults[queuenumber].update(
                self.__otherqueues[queuenumber].get())

    def __replacefuturesindict(self, dct):
        res = {}
        for (k, v) in dct.items():
            if isinstance(v, FutureResult) and v.creator in self.__otherqueues:
                res[k] = self.__getresultfromother(v.creator, v.id)
            elif isinstance(v, list):
                res[k] = self.__replacefuturesinlist(v)
            elif isinstance(v, tuple):
                res[k] = self.__replacefuturesintuple(v)
            elif isinstance(v, dict):
                res[k] = self.__replacefuturesindict(v)
            else:
                res[k] = v
        return res

    def __replacefuturesinlist(self, lst):
        res = []
        for e in lst:
            if isinstance(e, FutureResult) and e.creator in self.__otherqueues:
                res.append(self.__getresultfromother(e.creator, e.id))
            elif isinstance(e, list):
                res.append(self.__replacefuturesinlist(e))
            elif isinstance(e, tuple):
                res.append(self.__replacefuturesintuple(e))
            elif isinstance(e, dict):
                res.append(self.__replacefuturesindict(e))
            else:
                res.append(e)
        return res

    def __replacefuturesintuple(self, tpl):
        return tuple(self.__replacefuturesinlist(tpl))

    def __replacefuturesdirectly(self, args):
        for pos in self.__directupdates:
            if len(pos) == 2:
                x, y = pos
                fut = args[x][y]
                args[x][y] = self.__getresultfromother(fut.creator, fut.id)
            elif len(pos) == 3:
                x, y, z = pos
                fut = args[x][y][z]
                args[x][y][z] = self.__getresultfromother(fut.creator, fut.id)
            else:
                raise ValueError, 'Positions must be of length 2 or 3'

    def __decoupledworker(self):
        sys.excepthook = _getexcepthook()
        if hasattr(self._obj, '_decoupled') and callable(self._obj._decoupled):
            self._obj._decoupled()

        for (creatorid, queue) in self.__otherqueues.items():
            self.__otherresults[creatorid] = {}

        while True:
            batch = self.__toworker.get()
            ###
            if batch == 'STOP':
                self.__toworker.task_done()
                return
            ###
            resbatch = []
            for [id, funcname, args] in batch:
                if self.__otherqueues and args:
                    if self.__directupdates:
                        try:
                            self.__replacefuturesdirectly(args)
                        except KeyError:
                            args = self.__replacefuturesintuple(args)
                        except IndexError:
                            args = self.__replacefuturesintuple(args)
                    else:
                        args = self.__replacefuturesintuple(args)
                func = getattr(self._obj, funcname)
                res = func(*args) # NB: func's side-effects on args are ignored
                if id is not None:
                    resbatch.append((id, res)) 
            if self.__fromworker and resbatch:
                self.__fromworker.put(resbatch)
            self.__toworker.task_done()


    ### Stuff for the parent process

    def __getattr__(self, name):
        res = getattr(self._obj, name)
        if callable(res) and self.autowrap:
            def wrapperfunc(*args):
                return self._enqueue(name, *args)
            res = wrapperfunc
        setattr(self, name, res) # NB: Values are only read once...
        return res

    def _enqueue(self, funcname, *args):
        future = FutureResult(self.__instancenumber, self.__futurecnt)
        self.__futurecnt += 1
        self.__batch.append([future.id, funcname, args])
        if len(self.__batch) >= self.batchsize:
            self._endbatch()
        return future

    def _enqueuenoreturn(self, funcname, *args):
        self.__batch.append([None, funcname, args])
        if len(self.__batch) >= self.batchsize:
            self._endbatch()
        return None

    def _getresult(self, future):
        if self.__fromworker is None:
            raise RuntimeError, "Return values are not kept"
        if future.creator != self.__instancenumber:
            raise ValueError, "Cannot return results from other instances"
        # else find and return the result
        while True:
            if future.id in self.__results:
                return self.__results.pop(future.id)
            # else wait for results to become available
            self.__results.update(self.__fromworker.get())

    def _endbatch(self):
        if self.__batch:
            self.__toworker.put(self.__batch)
            self.__batch = []

    def _join(self):
        self._endbatch()
        self.__toworker.join()

    def shutdowndecoupled(self):
        """Let the Decoupled instance finish its tasks and stop it.

        The Decoupled instance should not be used after this.
        """
        self._join()
        self.__toworker.put('STOP')
        self.__toworker.join()
        self.__toworker.close()
        if self.__fromworker is not None:
            self.__fromworker.close()
        try:
            pygrametl._alltables.remove(self)
        except ValueError:
            pass


# SharedConnectionWrapper stuff
class SharedConnectionWrapperClient(object):
    """Provide access to a shared ConnectionWrapper.

    Users should not create a SharedConnectionWrapperClient directly, but
    instead use shareconnectionwrapper to do this.

    Each process should get its own SharedConnectionWrapper by calling
    the copy()/new() method.
    """

    def __init__(self, toserver, fromserver, freelines, connectionmodule,
                 userfuncnames=()):
        self.nametranslator = lambda s: s
        self.__clientid = None
        self.__toserver = toserver
        self.__fromserver = fromserver
        self.__freelines = freelines
        self.__connectionmodule = connectionmodule
        self.__userfuncnames = userfuncnames
        if pygrametl._defaulttargetconnection is None:
            pygrametl._defaulttargetconnection = self

    def __getstate__(self):
        res = self.__dict__.copy()
        res['_SharedConnectionWrapperClient__clientid'] = None
        return res

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.__createalluserfuncs() # A new self exists now

    def __del__(self):
        if self.__clientid is not None:
            self.__freelines.put(self.__clientid)

    def __connecttoSCWserver(self):
        self.__clientid = self.__freelines.get()

    def __enqueue(self, method, *args):
        if self.__clientid is None:
            self.__connecttoSCWserver()
        self.__toserver.put((self.__clientid, method, args))

    def __getrows(self, amount):
        # TODO:Should exceptions be transferred to the client and received here?
        self.__enqueue('#get', amount)
        return self.__fromserver[self.__clientid].get()

    def __join(self):
        self.__toserver.join()

    def __createalluserfuncs(self):
        for funcname in self.__userfuncnames:
            setattr(self, funcname, self.__createuserfunc(funcname))

    def __createuserfunc(self, funcname):
        def userfunction(*args):
            self.__enqueue('_userfunc_' + funcname, *args)
            # Wait for the userfunc to finish...
            res = self.__fromserver[self.__clientid].get() # OK after __enqueue
            assert res == 'USERFUNC'
        return userfunction

    def copy(self):
        """ Create a new copy of the SharedConnectionWrapper (same as new) """
        return copy.copy(self)

    def new(self):
        """ Create a new copy of the SharedConnectionWrapper (same as copy) """
        return self.copy()

    def execute(self, stmt, arguments=None, namemapping=None, translate=True):
        """Execute a statement.

           Arguments:
           - stmt: the statement to execute
           - arguments: a mapping with the arguments (default: None)
           - namemapping: a mapping of names such that if stmt uses %(arg)s
             and namemapping[arg]=arg2, the value arguments[arg2] is used 
             instead of arguments[arg]
           - translate: decides if translation from 'pyformat' to the
             undlying connection's format should take place. Default: True
        """
        if namemapping and arguments:
            arguments = pygrametl.copy(arguments, **namemapping)
        elif arguments:
            arguments = arguments.copy()
        self.__enqueue('execute', stmt, arguments, None, translate)

    def executemany(self, stmt, params, translate=True):
        """Execute a sequence of statements."""
        self.__enqueue('executemany', stmt, params, translate)

    def rowfactory(self, names=None):
        """Return a generator object returning result rows (i.e. dicts)."""
        (srvnames, rows) = self.__getrows(0)
        if names is None:
            names = srvnames
        for r in rows:
            yield dict(zip(names, r))

    def fetchone(self, names=None):
        """Return one result row (i.e. dict)."""
        (rownames, row) = self.__getrows(1)
        return dict(zip(names or rownames, row))

    def fetchonetuple(self):
        """Return one result tuple."""
        (rownames, row) = self.__getrows(1)
        return row

    def fetchmanytuples(self, cnt):
        """Return cnt result tuples."""
        (rownames, rows) = self.__getrows(cnt)
        return rows

    def fetchalltuples(self):
        """Return all result tuples"""
        (rownames, rows) = self.__getrows(0)
        return rows
        
    def rowcount(self):
        """Not supported. Returns -1."""
        return -1

    def getunderlyingmodule(self):
        """Return a reference to the underlying connection's module."""
        return self.__connectionmodule

    def commit(self):
        """Commit the transaction."""
        pygrametl.endload()
        self.__enqueue('commit')
        self.__join()

    def close(self):
        """Close the connection to the database,"""
        self.__enqueue('close')

    def rollback(self):
        """Rollback the transaction."""
        self.__enqueue('rollback')
        self.__join()

    def setasdefault(self):
        """Set this ConnectionWrapper as the default connection."""
        pygrametl._defaulttargetconnection = self

    def cursor(self):
        """Return a cursor object. Optional method."""
        raise NotImplementedError

    def resultnames(self):
        (rownames, nothing) = self.__getrows(None)
        return rownames


class SharedConnectionWrapperServer(object):
    """Manage access to a shared ConnectionWrapper.

    Users should not create a SharedConnectionWrapperServer directly, but
    instead use shareconnectionwrapper to do this.
    """

    def __init__(self, wrapped, toserver, toclients):
        self.__toserver = toserver
        self.__toclients = toclients
        self.__wrapped = wrapped
        self.__results = [(None, None) for q in toclients] #as (names, [tuples])

    def __senddata(self, client, amount=0):
        # Returns (column names, rows)
        # amount: None: No rows are returned - instead an empty list is sent
        #         0: all rows in a list, 
        #         1: a single row (NOT in a list),
        #         other positive numbers: max. that number of rows in a list. 
        (names, data) = self.__results[client]
        if amount is None:
            rows = []
        elif amount == 1 and data:
            rows = data.pop(0)
        elif amount > 0 and data:
            rows = data[0:amount]
            del data[0:amount]
        else:
            rows = data[:]
            del data[:]

        self.__toclients[client].put((names, rows))

    def worker(self):
        sys.excepthook = _getexcepthook()
        # TODO: Improved error handling such that an exception can be passed on 
        # to the responsible client. It is, however, likely that we cannot
        # continue using the shared DB connection after the exception occured...
        while True:
            (client, method, args) = self.__toserver.get()
            if method == '#get':
                self.__senddata(client, *args)
            elif method.startswith('_userfunc_'):
                target = getattr(self, method)
                target(*args)
                self.__toclients[client].put('USERFUNC')
            elif method == 'close':
                target = getattr(self.__wrapped, method)
                target(*args) # Probably no arguments anyway, but ...
            else: # it must be a function from the wrapped ConnectionWrapper
                target = getattr(self.__wrapped, method)
                target(*args)
                res = self.__wrapped.fetchalltuples()
                if not type(res) == list:
                    # In __senddata we pop/del from a list so a tuple won't work
                    res = list(res)
                self.__results[client] = (self.__wrapped.resultnames(), res)
            self.__toserver.task_done()
            if method == 'close':
                return


def shareconnectionwrapper(targetconnection, maxclients=10, userfuncs=()):
    """Share a ConnectionWrapper between several processes/threads.

    When Decoupled objects are used, they can try to update the DW at the same 
    time. They can use several ConnectionWrappers to avoid race conditions, but
    this is not transactionally safe. Instead, they can use a "shared" 
    ConnectionWrapper obtained through this function.

    When a ConnectionWrapper is shared, it is executing in a separate process
    (or thread, in case Jython is used) and ensuring that only one operation
    takes place at the time. This is hidden from the users of the shared 
    ConnectionWrapper.  They see an interface similar to the normal 
    ConnectionWrapper.

    When this method is called, it returns a SharedConnectionWrapperClient
    which can be used as a normal ConnectionWrapper. Each process 
    (i.e., each Decoupled object) should, however, get a unique
    SharedConnectionWrapperClient by calling copy() on the returned
    SharedConnectionWrapperClient.

    Note that a shared ConnectionWrapper needs to hold the complete result of 
    each query in memory until it is fetched by the process that executed the
    query. Again, this is hidden from the users.

    It is also possible to add methods to a shared ConnectionWrapper when it 
    is created. When this is done and the method is invoked, no other 
    operation will modify the DW at the same time. If, for example,
    the functions foo and bar are added to a shared ConnectionWrapper (by
    passing the argument userfuncs=(foo, bar) to shareconnectionwrapper),
    the returned SharedConnectionWrapperClient will offer the methods
    foo and bar which when called will be running in the separate process
    for the shared ConnectionWrapper. This is particularly useful for
    user-defined bulk loaders as used by BulkFactTable:

    def bulkload():
        # DBMS-specific code here.
        # No other DW operation should take place concurrently

    scw = shareconnectionwrapper(ConnectionWrapper(...), userfuncs=(bulkload,))
    facttbl = BulkFact(..., bulkloader=scw.copy().bulkload) #Note the .copy(). 

    Arguments:
    - targetconnection: a pygrametl ConnectionWrapper
    - maxclients: the maximum number of concurrent clients. Default: 10
    - userfuncs: a sequence of functions to add to the shared 
    ConnectionWrapper. Default: ()
    """
    toserver = multiprocessing.JoinableQueue(5000)
    toclients = [multiprocessing.Queue() for i in range(maxclients)]
    freelines = multiprocessing.Queue()
    for i in range(maxclients):
        freelines.put(i)
    serverCW = SharedConnectionWrapperServer(targetconnection, toserver, 
                                             toclients)
    userfuncnames = []
    for func in userfuncs:
        if not (callable(func) and hasattr(func, 'func_name') and \
                    not func.func_name == '<lambda>'):
            raise ValueError, "Elements in userfunc must be callable and named"
        if hasattr(SharedConnectionWrapperClient, func.func_name):
            raise ValueError, "Illegal function name: " + func.func_name
        setattr(serverCW, '_userfunc_' + func.func_name, func)
        userfuncnames.append(func.func_name)
    serverprocess = multiprocessing.Process(target=serverCW.worker)
    serverprocess.name = 'Process for shared connection wrapper'
    serverprocess.daemon = True
    serverprocess.start()
    module = targetconnection.getunderlyingmodule()
    clientCW = SharedConnectionWrapperClient(toserver, toclients, freelines,
                                             module, userfuncnames)
    return clientCW


# Shared sequences
def getsharedsequencefactory(startvalue, intervallen=5000):
    """ Creates a factory for parallel readers of a sequence.
    
        Returns a callable f. When f() is called, it returns a callable g.
        Whenever g(*args) is called, it returns a unique int from a sequence
        (if several g's are created, the order of the calls may lead to that
        the returned ints are not ordered, but they will be unique). The
        arguments to g are ignored, but accepted. Thus g can be used as
        idfinder for [Decoupled]Dimensions.

        The different g's can be used safely from different processes and 
        threads.

        Arguments:
        - startvalue: The first value to return. If None, 0 is assumed.
        - intervallen: The amount of numbers that a single g from above
          can return before synchronization is needed to get a new amount.
          Default: 5000.
    """
    if startvalue is None:
        startvalue = 0

    # We use a Queue to ensure that intervals are only given to one deliverer
    values = multiprocessing.Queue(10)

    # A worker that fills the queue
    def valuegenerator(nextval):
        sys.excepthook = _getexcepthook()
        while True:
            values.put((nextval, nextval + intervallen))
            nextval += intervallen

    p = multiprocessing.Process(target=valuegenerator, args=(startvalue,))
    p.daemon = True
    p.start()

    # A generator that repeatedly gets an interval from the queue and returns 
    # all numbers in that interval before it gets a new interval and goes on ...
    def valuedeliverer():
        while True:
            interval = values.get()
            for i in range(*interval):
                yield i

    # A factory method for the object the end-consumer calls
    def factory():
        generator = valuedeliverer() # get a unique generator
        # The method called (i.e., the g) by the end-consumer
        def getnextseqval(*ignored):
            return generator.next()
        return getnextseqval

    return factory


# -*- coding: utf-8 -*-
""" A module for Jython emulating (a small part of) CPython's multiprocessing.
    With this, pygrametl can be made to use multiprocessing, but actually use       threads when used from Jython (where there is no GIL).
"""

__author__ = "Christian Thomsen"
__maintainer__ = "Sílex Sistemas Ltda."
__version__ = '0.0.1'
__all__ = ['JoinableQueue', 'Process', 'Queue', 'Value']

import sys
if not sys.platform.startswith('java'):
    raise ImportError, 'jythonmultiprocessing is made for Jython'

from threading import Thread
from Queue import Queue
from pygrametl.jythonsupport import Value


class Process(Thread):
    pid = '<n/a>'
    daemon = property(Thread.isDaemon, Thread.setDaemon)
    name = property(Thread.getName, Thread.setName)


class JoinableQueue(Queue):
    def close(self):
        pass


# -*- coding: utf-8 -*-
"""This module contains classes for looking up rows, inserting rows
   and updating rows in dimensions and fact tables. Rows are represented
   as dictionaries mapping between attribute names and attribute values.

   Many of the class methods take an optional 'namemapping' argument which is
   explained here, but not repeated in the documentation for the individual
   methods: Consider a method m which is given a row r and a namemapping n.
   Assume that the method m uses the attribute a in r (i.e., r[a]). If the
   attribute a is not in the namemapping, m will just use r[a] as expected.
   But if the attribute a is in the namemapping, the name a is mapped to
   another name and the other name is used. That means that m then uses
   r[n[a]].  This is practical if attribute names in the considered rows and
   DW tables differ. If, for example, data is inserted into an order dimension
   in the DW that has the attribute order_date, but the source data uses the
   attribte name date, we can use a name mapping from order_date to date:
   dim.insert(row=..., namemapping={'order_date':'date'})
"""

from subprocess import Popen, PIPE
from time import sleep
import types
import tempfile

import pygrametl
import pygrametl.parallel
from pygrametl.FIFODict import FIFODict

__author__ = "Christian Thomsen"
__maintainer__ = "Sílex Sistemas Ltda."
__version__ = '0.0.1'
__all__ = ['Dimension', 'CachedDimension', 'SlowlyChangingDimension',
           'SnowflakedDimension', 'FactTable', 'BatchFactTable',
           'BulkFactTable', 'SubprocessFactTable', 'DecoupledDimension',
           'DecoupledFactTable', 'BasePartitioner', 'DimensionPartitioner',
           'FactTablePartitioner']


class Dimension(object):
    """A class for accessing a dimension. Does no caching."""

    def __init__(self, name, key, attributes, lookupatts=(), 
                 idfinder=None, defaultidvalue=None, rowexpander=None,
                targetconnection=None):
        """Arguments:
           - name: the name of the dimension table in the DW
           - key: the name of the primary key in the DW
           - attributes: a sequence of the attribute names in the dimension 
             table. Should not include the name of the primary key which is
             given in the key argument.
           - lookupatts: A subset of the attributes that uniquely identify
             a dimension members. These attributes are thus used for looking
             up members. If not given, it is assumed that 
             lookupatts = attributes
           - idfinder: A function(row, namemapping) -> key value that assigns
             a value to the primary key attribute based on the content of the
             row and namemapping. If not given, it is assumed that the primary
             key is an integer, and the assigned key value is then the current
             maximum plus one.
           - defaultidvalue: An optional value to return when a lookup fails. 
             This should thus be the ID for a preloaded "Unknown" member.
           - rowexpander: A function(row, namemapping) -> row. This function
             is called by ensure before insertion if a lookup of the row fails.
             This is practical if expensive calculations only have to be done
             for rows that are not already present. For example, for a date
             dimension where the full date is used for looking up rows, a
             rowexpander can be set such that week day, week number, season, 
             year, etc. are only calculated for dates that are not already
             represented. If not given, no automatic expansion of rows is
             done.
           - targetconnection: The ConnectionWrapper to use. If not given,
             the default target connection is used.
        """
        if not type(key) in types.StringTypes:
            raise ValueError, "Key argument must be a string"
        if not len(attributes):
            raise ValueError, "No attributes given"

        if targetconnection is None:
            targetconnection = pygrametl.getdefaulttargetconnection()
            if targetconnection is None:
                raise ValueError, "No target connection available"
        self.targetconnection = targetconnection
        self.name = name
        self.attributes = attributes
        self.key = key
        if lookupatts == ():
            lookupatts = attributes
        self.all = [key,]
        self.all.extend(attributes)
        self.lookupatts = lookupatts
        self.defaultidvalue = defaultidvalue
        self.rowexpander = rowexpander
        pygrametl._alltables.append(self)

        # Now create the SQL that we will need...

        # This gives "SELECT key FROM name WHERE lookupval1 = %(lookupval1)s 
        #             AND lookupval2 = %(lookupval2)s AND ..."
        self.keylookupsql = "SELECT " + key + " FROM " + name + " WHERE " + \
            " AND ".join(["%s = %%(%s)s" % (lv, lv) for lv in lookupatts])

        # This gives "SELECT key, att1, att2, ... FROM NAME WHERE key = %(key)s"
        self.rowlookupsql = "SELECT " + ", ".join(self.all) +  \
            " FROM %s WHERE %s = %%(%s)s" % (name, key, key)

        # This gives "INSERT INTO name(key, att1, att2, ...) 
        #             VALUES (%(key)s, %(att1)s, %(att2)s, ...)"
        self.insertsql = "INSERT INTO " + name + "(%s" % (key,) + \
            (attributes and ", " or "") + \
            ", ".join(attributes) + ") VALUES (" + \
            ", ".join(["%%(%s)s" % (att,) for att in self.all]) + ")"

        if idfinder is not None:
            self.idfinder = idfinder
        else:
            self.targetconnection.execute("SELECT MAX(%s) FROM %s" % \
                                              (key, name))
            self.__maxid = self.targetconnection.fetchonetuple()[0]
            if self.__maxid is None:
                self.__maxid = 0
            self.idfinder = self._getnextid

    def lookup(self, row, namemapping={}):
        """ Find the key for the row with the given values.

            Arguments:
            - row: a dict which must contain at least the lookup attributes
            - namemapping: an optional namemapping (see module's documentation)
        """
        key = self._before_lookup(row, namemapping)
        if key is not None:
            return key
        
        self.targetconnection.execute(self.keylookupsql, row, namemapping)
        
        keyvalue = self.targetconnection.fetchonetuple()[0]
        if keyvalue is None:
            keyvalue = self.defaultidvalue  # most likely also None...

        self._after_lookup(row, namemapping, keyvalue)
        return keyvalue

    def _before_lookup(self, row, namemapping):
        return None

    def _after_lookup(self, row, namemapping, resultkeyvalue):
        pass

    def getbykey(self, keyvalue):
        """Lookup and return the row with the given key value.

           If no row is found in the dimension table, the function returns
           a row where all values (including the key) are None.
        """
        if type(keyvalue) == types.DictType:
            keyvalue = keyvalue[self.key]

        row = self._before_getbykey(keyvalue)
        if row is not None:
            return row
        self.targetconnection.execute(self.rowlookupsql, {self.key : keyvalue})
        row = self.targetconnection.fetchone(self.all)
        self._after_getbykey(keyvalue, row)
        return row

    def _before_getbykey(self, keyvalue):
        return None

    def _after_getbykey(self, keyvalue, resultrow):
        pass

    def getbyvals(self, values, namemapping={}):
        """Return a list of all rows with values identical to the given.

           Arguments:
           - values: a dict which must hold a subset of the tables attributes.
             All rows that have identical values for all attributes in this
             dict are returned.
           - namemapping: an optional namemapping (see module's documentation)
        """
        res = self._before_getbyvals(values, namemapping)
        if res is not None:
            return res

        # select all attributes from the table. The attributes available from 
        # the values dict are used in the WHERE clause.
        attstouse = [a for a in self.attributes \
                         if a in values or a in namemapping]
        sql = "SELECT " + ", ".join(self.all) + " FROM " + self.name + \
            " WHERE " + \
            " AND ".join(["%s = %%(%s)s" % (att, att) for att in attstouse])

        self.targetconnection.execute(sql, values, namemapping)

        res = [r for r in self.targetconnection.rowfactory(self.all)]
        self._after_getbyvals(values, namemapping, res)
        return res

    def _before_getbyvals(self, values, namemapping):
        return None

    def _after_getbyvals(self, values, namemapping, resultrows):
        pass

    def update(self, row, namemapping={}):
        """Update a single row in the dimension table.

           Arguments:
            - row: a dict which must contain the key for the dimension.
              The row with this key value is updated such that it takes
              the value of row[att] for each attribute att which is also in 
              row.
            - namemapping: an optional namemapping (see module's documentation)
        """
        res = self._before_update(row, namemapping)
        if res:
            return

        if self.key not in row:
            raise KeyError, "The key value (%s) is missing in the row" % \
                (self.key,)

        attstouse = [a for a in self.attributes \
                         if a in row or a in namemapping]
        if not attstouse:
            # Only the key was there - there are no attributes to update
            return

        sql = "UPDATE " + self.name + " SET " + \
            ", ".join(["%s = %%(%s)s" % (att, att) for att in attstouse]) + \
            " WHERE %s = %%(%s)s" % (self.key, self.key)
        self.targetconnection.execute(sql, row, namemapping)
        self._after_update(row, namemapping)

    def _before_update(self, row, namemapping):
        return None

    def _after_update(self, row, namemapping):
        pass

    def ensure(self, row, namemapping={}):
        """Lookup the given row. If that fails, insert it. Return the key value.

           If the lookup fails and a rowexpander was set when creating the
           instance, this rowexpander is called before the insert takes place.

           Arguments:
           - row: the row to lookup or insert. Must contain the lookup 
             attributes. 
           - namemapping: an optional namemapping (see module's documentation)
        """
        res = self.lookup(row, namemapping)
        if res is not None:
            return res
        else:
            if self.rowexpander:
                row = self.rowexpander(row, namemapping)
            return self.insert(row, namemapping)

    def insert(self, row, namemapping={}):
        """Insert the given row. Return the new key value.

           Arguments:
           - row: the row to insert. The dict is not updated. It must contain
             all attributes, and is allowed to contain more attributes than
             that.
           - namemapping: an optional namemapping (see module's documentation)
        """
        res = self._before_insert(row, namemapping)
        if res is not None:
            return res

        key = (namemapping.get(self.key) or self.key)
        if row.get(key) is None:
            keyval = self.idfinder(row, namemapping)
            row[key] = keyval
            keyadded = True
        else:
            keyval = row[key]
            keyadded = False

        self.targetconnection.execute(self.insertsql, row, namemapping)
        if keyadded:
            del row[key]
        self._after_insert(row, namemapping, keyval)
        return keyval

    def _before_insert(self, row, namemapping):
        return None

    def _after_insert(self, row, namemapping, newkeyvalue):
        pass


    def _getnextid(self, ignoredrow, ignoredmapping):
        self.__maxid += 1
        return self.__maxid

    def endload(self):
        """Finalize the load."""
        pass


class CachedDimension(Dimension):
    """A class for accessing a dimension. Does caching.

       We assume that the DB doesn't change or add any attribute
       values that are cached.
       For example, a DEFAULT value in the DB can break this assumption.
    """

    def __init__(self, name, key, attributes, lookupatts=(), 
                 idfinder=None, defaultidvalue=None, rowexpander=None,
                 size=10000, prefill=False, cachefullrows=False,
                 cacheoninsert=True, targetconnection=None):
        """Arguments:
           - name: the name of the dimension table in the DW
           - key: the name of the primary key in the DW
           - attributes: a sequence of the attribute names in the dimension 
             table. Should not include the name of the primary key which is
             given in the key argument.
           - lookupatts: A subset of the attributes that uniquely identify
             a dimension members. These attributes are thus used for looking
             up members. If not given, it is assumed that 
             lookupatts = attributes
           - idfinder: A function(row, namemapping) -> key value that assigns
             a value to the primary key attribute based on the content of the
             row and namemapping. If not given, it is assumed that the primary
             key is an integer, and the assigned key value is then the current
             maximum plus one.
           - defaultidvalue: An optional value to return when a lookup fails. 
             This should thus be the ID for a preloaded "Unknown" member.
           - rowexpander: A function(row, namemapping) -> row. This function
             is called by ensure before insertion if a lookup of the row fails.
             This is practical if expensive calculations only have to be done
             for rows that are not already present. For example, for a date
             dimension where the full date is used for looking up rows, a
             rowexpander can be set such that week day, week number, season, 
             year, etc. are only calculated for dates that are not already
             represented. If not given, no automatic expansion of rows is
             done.
           - size: the maximum number of rows to cache. If less than or equal
             to 0, unlimited caching is used. Default: 10000
           - prefill: a flag deciding if the cache should be filled when
             initialized. Default: False
           - cachefullrows: a flag deciding if full rows should be
             cached. If not, the cache only holds a mapping from
             lookupattributes to key values. Default: False.
           - cacheoninsert: a flag deciding if the cache should be updated
             when insertions are done. Default: True
           - targetconnection: The ConnectionWrapper to use. If not given,
             the default target connection is used.
        """

        Dimension.__init__(self, name, key, attributes, lookupatts, idfinder, 
                           defaultidvalue, rowexpander, targetconnection)
        self.cacheoninsert = cacheoninsert
        self.__prefill = prefill
        self.__size = size
        if size > 0:
            if cachefullrows:
                self.__key2row = FIFODict(size)
            self.__vals2key = FIFODict(size)
        else:
            # Use dictionaries as unlimited caches
            if cachefullrows:
                self.__key2row = {}
            self.__vals2key = {}

        self.cachefullrows = cachefullrows

        if prefill:
            if cachefullrows:
                positions = tuple([self.all.index(att) \
                                       for att in self.lookupatts])
                # select the key and all attributes
                sql = "SELECT %s FROM %s" % (", ".join(self.all), name)
            else:
                # select the key and the lookup attributes
                sql = "SELECT %s FROM %s" % \
                    (", ".join([key] + [l for l in self.lookupatts]), name)
                positions = range(1, len(self.lookupatts) + 1)

            self.targetconnection.execute(sql)
            if size <= 0:
                data = self.targetconnection.fetchalltuples()
            else:
                data = self.targetconnection.fetchmanytuples(size)

            for rawrow in data:
                if cachefullrows:
                    self.__key2row[rawrow[0]] = rawrow
                t = tuple([rawrow[i] for i in positions])
                self.__vals2key[t] = rawrow[0]

    def lookup(self, row, namemapping={}):
        if self.__prefill and self.cacheoninsert and \
                (self.__size <= 0 or len(self.__vals2key) < self.__size):
            # Everything is cached. We don't have to look in the DB
            res = self._before_lookup(row, namemapping)
            if res is not None:
                return res
            else:
                return self.defaultidvalue
        else:
            # Something is not cached so we have to use the classical lookup
            return Dimension.lookup(self, row, namemapping)

    def _before_lookup(self, row, namemapping):
        namesinrow =[(namemapping.get(a) or a) for a in self.lookupatts]
        searchtuple = tuple([row[n] for n in namesinrow])
        return self.__vals2key.get(searchtuple, None)

    def _after_lookup(self, row, namemapping, resultkey):
        if resultkey is not None:
            namesinrow =[(namemapping.get(a) or a) for a in self.lookupatts]
            searchtuple = tuple([row[n] for n in namesinrow])
            self.__vals2key[searchtuple] = resultkey

    def _before_getbykey(self, keyvalue):
        if self.cachefullrows:
            res = self.__key2row.get(keyvalue)
            if res is not None:
                return dict(zip(self.all, res))
        return None

    def _after_getbykey(self, keyvalue, resultrow):
        if self.cachefullrows and resultrow[self.key] is not None:
            # if resultrow[self.key] is None, no result was found in the db
            self.__key2row[keyvalue] = tuple([resultrow[a] for a in self.all])

    def _before_update(self, row, namemapping):
        """ """
        # We have to remove old values from the caches.
        key = (namemapping.get(self.key) or self.key)
        for att in self.lookupatts:
            if ((att in namemapping and namemapping[att] in row) or att in row):
                # A lookup attribute is about to be changed and we should make
                # sure that the cache does not map from the old value.  Here,
                # we can only see the new value, but we can get the old lookup
                # values by means of the key:
                oldrow = self.getbykey(row[key])
                namesinrow =[(namemapping.get(a) or a) for a in self.lookupatts]
                searchtuple = tuple([oldrow[n] for n in namesinrow])
                if searchtuple in self.__vals2key:
                    del self.__vals2key[searchtuple]
                break
                

        if self.cachefullrows:
            if row[key] in self.__key2row:
                # The cached row is now incorrect. We must make sure it is
                # not in the cache.
                del self.__key2row[row[key]]

        return None

    def _after_insert(self, row, namemapping, newkeyvalue):
        """ """
        # After the insert, we can look the row up. Pretend that we 
        # did that. Then we get the new data cached.
        # NB: Here we assume that the DB doesn't change or add anything.
        # For example, a DEFAULT value in the DB breaks this assumption.
        if self.cacheoninsert:
            self._after_lookup(row, namemapping, newkeyvalue)
            if self.cachefullrows:
                tmp = pygrametl.project(self.all, row, namemapping)
                tmp[self.key] = newkeyvalue
                self._after_getbykey(newkeyvalue, tmp)


class SlowlyChangingDimension(Dimension):
    """A class for accessing a slowly changing dimension of "type 2".
    
       "Type 1" updates can also be applied for a subset of the attributes.

       Caching is used. We assume that the DB doesn't change or add any 
       attribute values that are cached. For example, a DEFAULT value in the 
       DB can break this assumption.
    """

    def __init__(self, name, key, attributes, lookupatts, versionatt, 
                 fromatt=None, fromfinder=None,
                 toatt=None, tofinder=None, minfrom=None, maxto=None,
                 srcdateatt=None, srcdateparser=pygrametl.ymdparser,
                 type1atts=(), cachesize=10000, prefill=False, idfinder=None,
                 targetconnection=None):
        """Arguments:
           - name: the name of the dimension table in the DW
           - key: the name of the primary key in the DW
           - attributes: a sequence of the attribute names in the dimension 
             table. Should not include the name of the primary key which is
             given in the key argument, but should include versionatt,
             fromatt, and toatt.
           - lookupatts: a sequence with a subset of the attributes that 
             uniquely identify a dimension members. These attributes are thus 
             used for looking up members. If not given, it is assumed that 
             lookupatts = attributes
           - versionatt: the name of the attribute holding the version number
           - fromatt: the name of the attribute telling from when the version
             becomes valid. Not used if None. Default: None
           - fromfinder: a function(targetconnection, row, namemapping) 
             returning a value for the fromatt for a new version (the function
             is first used when it is determined that a new version must be
             added; it is not applied to determine this). 
             If fromfinder is None and srcdateatt is also None, 
             pygrametl.today is used as fromfinder. If fromfinder is None 
             and srcdateatt is not None, 
             pygrametl.datereader(srcdateatt, srcdateparser) is used. 
             In other words, if no date attribute and no special
             date function are given, new versions get the date of the current
             day. If a date attribute is given (but no date function), the
             date attribute's value is converted (by means of srcdateparser)
             and a new version gets the result of this as the date it is valid
             from. Default: None
           - toatt: the name of the attribute telling until when the version
             is valid. Not used if None. Default: None
           - tofinder: a function(targetconnection, row, namemapping)
             returning a value for the toatt. If not set, fromfinder is used
             (note that if fromfinder is None, it is set to a default
             function -- see the comments about fromfinder. The possibly
             modified value is used here.) Default: None
           - minfrom: the value to use for fromatt for the 1st version of a 
             member if fromatt is not already set. If None, the value is 
             found in the same way as for other new versions, i.e., as 
             described for fromfinder. If fromatt should take the value
             NULL for the 1st version, set minfrom to a tuple holding a single 
             element which is None: (None,). Note that minto affects the 1st 
             version, not any following versions. Note also that if the member 
             to insert already contains a value for fromatt, minfrom is ignored.
             Default: None.
             ADDED IN THE AUGUST 2013 RELEASE. The minfrom argument is not 
             available in previous releases.
           - maxto: the value to use for toatt for new members. Default: None
           - srcdateatt: the name of the attribute in the source data that
             holds a date showing when a version is valid from. The data is
             converted to a datetime by applying srcdateparser on it.
             If not None, the date attribute is also used when comparing
             a potential new version to the newest version in the DB.
             If None, the date fields are not compared. Default: None
           - srcdateparser: a function that takes one argument (a date in the
             format scrdateatt has) and returns a datetime.datetime.
             If srcdateatt is None, srcdateparser is not used.
             Default: pygrametl.ymdparser (i.e., the default value is a
             function that parses a string of the form 'yyyy-MM-dd')
           - type1atts: a sequence of attributes that should have type1 updates
             applied. Default: ()
           - cachesize: the maximum size of the cache. 0 disables caching
             and values smaller than 0 allows unlimited caching
           - prefill: decides if the cache should be prefilled with the newest
             versions. Default: False. NB: This is a new argument in ver. 0.2.0.
           - idfinder: a function(row, namemapping) -> key value that assigns
             a value to the primary key attribute based on the content of the
             row and namemapping. If not given, it is assumed that the primary
             key is an integer, and the assigned key value is then the current
             maximum plus one.
           - targetconnection: The ConnectionWrapper to use. If not given,
             the default target connection is used.
        """
        # TODO: Should scdensure just override ensure instead of being a new 
        #       method?

        Dimension.__init__(self, name, key, attributes, lookupatts, 
                           idfinder, None, None, targetconnection)

        if not versionatt:
            raise ValueError, 'A version attribute must be given'

        self.versionatt = versionatt
        self.fromatt = fromatt
        if fromfinder is not None:
            self.fromfinder = fromfinder
        elif srcdateatt is not None: #and fromfinder is None
            self.fromfinder = pygrametl.datereader(srcdateatt, srcdateparser)
        else:                        #fromfinder is None and srcdateatt is None
            self.fromfinder = pygrametl.today
        self.toatt = toatt
        if tofinder is None:
            tofinder = self.fromfinder
        self.tofinder = tofinder
        self.minfrom = minfrom
        self.maxto = maxto
        self.srcdateatt = srcdateatt
        self.srcdateparser = srcdateparser
        self.type1atts = type1atts
        if cachesize > 0:
            self.rowcache = FIFODict(cachesize)
            self.keycache = FIFODict(cachesize)
        elif cachesize < 0:
            self.rowcache = {}
            self.keycache = {}
        #else cachesize == 0 and we do not create any caches
        self.__cachesize = cachesize
        self.__prefill = prefill

        # Check that versionatt, fromatt and toatt are also declared as 
        # attributes
        for var in (versionatt, fromatt, toatt):
            if var and var not in attributes:
                raise ValueError, "%s not present in attributes argument" % \
                    (var,)

        # Now extend the SQL from Dimension such that we use the versioning
        self.keylookupsql += " ORDER BY %s DESC" % (versionatt,)

        if toatt:
            self.updatetodatesql = \
                "UPDATE %s SET %s = %%(%s)s WHERE %s = %%(%s)s" % \
                (name, toatt, toatt, key, key)

        if prefill:
            self.__prefillcaches()

    def __prefillcaches(self):
        args = None
        if self.toatt:
            # We can use the toatt to see if rows are still current.
            # Select all attributes from the rows where maxto is set to the
            # default value (which may be NULL)
            sql = 'SELECT %s FROM %s WHERE %s %s' % \
                (', '.join(self.all), self.name, self.toatt, 
                 self.maxto is None and 'IS NULL' or '= %(maxto)s')
            if self.maxto is not None:
                args = {'maxto':self.maxto}
        else:
            # We have to find max(versionatt) for each group of lookupatts and
            # do a join to get the right rows.
            lookupattlist = ', '.join(self.lookupatts)
            newestversions = 'SELECT %s, MAX(%s) AS %s FROM %s GROUP BY %s' % \
                (lookupattlist, self.versionatt, self.versionatt, self.name, \
                     lookupattlist)
            joincond = ' AND '.join(['A.%s = B.%s' % (att, att) for att in \
                                         [l for l in self.lookupatts] + \
                                         [self.versionatt]
                                     ])
            sql = 'SELECT %s FROM (%s) AS A, %s AS B WHERE %s' %\
                (', '.join(['B.%s AS %s' % (att, att) for att in self.all]), 
                 newestversions, self.name, joincond)
            
        # sql is a statement that fetches the newest versions. Fill the caches
        positions = [self.all.index(att) for att in self.lookupatts]
        self.targetconnection.execute(sql, args)
        for rawrow in self.targetconnection.fetchmanytuples(self.__cachesize):
            self.rowcache[rawrow[0]] = rawrow
            t = tuple([rawrow[i] for i in positions])
            self.keycache[t] = rawrow[0]

    def lookup(self, row, namemapping={}):
        """ Find the key for the newest version with the given values.

            Arguments:
            - row: a dict which must contain at least the lookup attributes
            - namemapping: an optional namemapping (see module's documentation)
        """
        if self.__prefill and (self.__cachesize < 0 or \
                                   len(self.keycache) < self.__cachesize):
            # Everything is cached. We don't have to look in the DB
            return self._before_lookup(row, namemapping)
        else:
            # Something is not cached so we have to use the classical lookup.
            # Note that __init__ updated self.keylookupsql to use ORDER BY ...
            return Dimension.lookup(self, row, namemapping)

    def scdensure(self, row, namemapping={}):
        """Lookup or insert a version of a slowly changing dimension member.

           NB: Has side-effects on the given row.

           Arguments:
           - row: a dict containing the attributes for the member. 
             key, versionatt, fromatt, and toatt are not required to be
             present but will be added (if defined).
           - namemapping: an optional namemapping (see module's documentation)
        """
        versionatt = (namemapping.get(self.versionatt) or self.versionatt)
        key = (namemapping.get(self.key) or self.key)
        if self.fromatt: # this protects us against None in namemapping.
            fromatt = (namemapping.get(self.fromatt) or self.fromatt)
        else:
            fromatt = None
        if self.toatt:
            toatt = (namemapping.get(self.toatt) or self.toatt)
        else:
            toatt = None
        if self.srcdateatt:
            srcdateatt = (namemapping.get(self.srcdateatt) or self.srcdateatt)
        else:
            srcdateatt = None

        # Get the newest version and compare to that
        keyval = self.lookup(row, namemapping)
        if keyval is None:
            # It is a new member. We add the first version.
            row[versionatt] = 1
            if fromatt and fromatt not in row:
                if self.minfrom is not None:
                    # We need the following hack to distinguish between
                    # 'not set' and 'use the value None'...
                    if self.minfrom == (None,): row[fromatt] = None
                    else: row[fromatt] = self.minfrom
                else:
                    row[fromatt] = self.fromfinder(self.targetconnection, 
                                                   row, namemapping)
            if toatt and toatt not in row:
                row[toatt] = self.maxto
            row[key] = self.insert(row, namemapping)
            return row[key]
        else:
            # There is an existing version. Check if the attributes are 
            # identical
            type1updates = {} # for type 1
            addnewversion = False # for type 2
            other = self.getbykey(keyval) # the full existing version
            for att in self.all:
                # Special (non-)handling of versioning and key attributes:
                if att in (self.key, self.versionatt, self.toatt):
                    # Don't compare these - we don't expect them to have 
                    # meaningful values in row
                    continue
                # We may have to compare the "from dates"
                elif att == self.fromatt:
                    if self.srcdateatt is None: # We don't compare dates then
                        continue
                    else:
                        # We have to compare the dates in row[..] and other[..].
                        # We have to make sure that the dates are of comparable
                        # types.
                        rdt = self.srcdateparser(row[srcdateatt])
                        if rdt == other[self.fromatt]:
                            continue # no change in the "from attribute"
                        elif type(rdt) == type(other[self.fromatt]):
                            # they are not equal but are of the same type, so we
                            # are dealing with a new date
                            addnewversion = True
                        else:
                            # They have different types (and are thus not
                            # equal). Try to convert to strings and see if they
                            # are equal.
                            modref = self.targetconnection.getunderlyingmodule()
                            rowdate = modref.Date(rdt.year, rdt.month, rdt.day)
                            if str(rowdate).strip('\'"') != \
                                    str(other[self.fromatt]).strip('\'"'):
                                addnewversion = True
                # Handling of "normal" attributes:
                else:
                    mapped = (namemapping.get(att) or att)
                    if row[mapped] != other[att]:
                        if att in self.type1atts:
                            type1updates[att] = row[mapped]
                        else:
                            addnewversion = True
                if addnewversion and not self.type1atts:
                    # We don't have to look for possible type 1 updates
                    # and we already know that a type 2 update is needed.
                    break
                #else: continue

            if len(type1updates) > 0:
                # Some type 1 updates were found
                self.__performtype1updates(type1updates, other)
            
            if addnewversion: # type 2
                # Make a new row version and insert it
                row.pop(key, None)
                row[versionatt] = other[self.versionatt] + 1
                if fromatt:
                    row[fromatt] = self.fromfinder(self.targetconnection, 
                                                   row, namemapping)
                if toatt:
                    row[toatt] = self.maxto
                row[key] = self.insert(row, namemapping)
                # Update the todate attribute in the old row version in the DB.
                if toatt:
                    toattval = self.tofinder(self.targetconnection, row, 
                                             namemapping)
                    self.targetconnection.execute(self.updatetodatesql, 
                                     {self.key : keyval, self.toatt : toattval})
                # Only cache the newest version - this is new in ver. 0.2.0!
                if keyval in self.rowcache:
                    del self.rowcache[keyval]
            else:
                # Update the row dict by giving version and dates and the key
                row[key] = keyval
                row[versionatt] = other[self.versionatt]
                if self.fromatt:
                    row[fromatt] = other[self.fromatt]
                if self.toatt:
                    row[toatt] = other[self.toatt]

            return row[key]

    def _before_lookup(self, row, namemapping):
        if self.__cachesize:
            namesinrow =[(namemapping.get(a) or a) for a in self.lookupatts]
            searchtuple = tuple([row[n] for n in namesinrow])
            return self.keycache.get(searchtuple, None)
        return None

    def _after_lookup(self, row, namemapping, resultkey):
        if self.__cachesize and resultkey is not None:
            namesinrow =[(namemapping.get(a) or a) for a in self.lookupatts]
            searchtuple = tuple([row[n] for n in namesinrow])
            self.keycache[searchtuple] = resultkey

    def _before_getbykey(self, keyvalue):
        if self.__cachesize:
            res = self.rowcache.get(keyvalue)
            if res is not None:
                return dict(zip(self.all, res))
        return None

    def _after_getbykey(self, keyvalue, resultrow):
        if self.__cachesize and resultrow[self.key] is not None:
            # if resultrow[self.key] is None, no result was found in the db
            self.rowcache[keyvalue] = tuple([resultrow[a] for a in self.all])

    def _before_update(self, row, namemapping):
        """ """
        # We have to remove old values from the caches.
        key = (namemapping.get(self.key) or self.key)
        for att in self.lookupatts:
            if (att in namemapping or att in row):
                # A lookup attribute is about to be changed and we should make
                # sure that the cache does not map from the old value.  Here,
                # we can only see the new value, but we can get the old lookup
                # values by means of the key:
                oldrow = self.getbykey(row[key])
                namesinrow =[(namemapping.get(a) or a) for a in self.lookupatts]
                searchtuple = tuple([oldrow[n] for n in namesinrow])
                if searchtuple in self.keycache:
                    del self.keycache[searchtuple]
                break

        if row[key] in self.rowcache:
            # The cached row is now incorrect. We must make sure it is
            # not in the cache.
            del self.rowcache[row[key]]

        return None

    def _after_insert(self, row, namemapping, newkeyvalue):
        """ """
        # After the insert, we can look it up. Pretend that we 
        # did that. Then we get the new data cached.
        # NB: Here we assume that the DB doesn't change or add anything.
        # For example, a DEFAULT value in the DB breaks this assumption.
        # Note that we always cache inserted members (in CachedDimension
        # this is an option).
        if self.__cachesize:
            self._after_lookup(row, namemapping, newkeyvalue)
            tmp = pygrametl.project(self.all[1:], row, namemapping)
            tmp[self.key] = newkeyvalue
            self._after_getbykey(newkeyvalue, tmp)

    def __performtype1updates(self, updates, lookupvalues, namemapping={}):
        """ """
        # find the keys in the rows that should be updated
        self.targetconnection.execute(self.keylookupsql, lookupvalues, 
                                      namemapping)
        updatekeys = [e[0] for e in self.targetconnection.fetchalltuples()]
        updatekeys.reverse()
        # Generate SQL for the update
        valparts = ", ".join(["%s = %%(%s)s" % (k, k) for k in updates])
        keyparts = ", ".join([str(k) for k in updatekeys])
        sql = "UPDATE %s SET %s WHERE %s IN (%s)" % \
            (self.name, valparts, self.key, keyparts)
        self.targetconnection.execute(sql, updates)
        # Remove from our own cache
        for key in updatekeys:
            if key in self.rowcache:
                del self.rowcache[key]
        
                
SCDimension = SlowlyChangingDimension


# NB: SnowflakedDimension's methods may have side-effects:
# row[somedim.key] = someval.

class SnowflakedDimension(object):
    """A class for accessing a snowflaked dimension spanning several tables
       in the underlying database. Lookups and inserts are then automatically
       spread out over the relevant tables while the programmer only needs
       to interact with a single SnowflakedDimension instance.
    """

    def __init__(self, references, expectboguskeyvalues=False):
        """Arguments:
           - references: a sequence of pairs of Dimension objects 
             [(a1,a2), (b1,b2), ...] meaning that a1 has a foreign key to a2 
             etc. a2 may itself be a sequence of Dimensions:
             [(a1, [a21, a22, ...]), (b1, [b21, b22, ...]), ...].

             The first element of the first pair (a1 in the example above) must
             be the dimension table representing the lowest level in the 
             hierarchy (i.e., the dimension table the closest to the fact 
             table).

             Each dimension must be reachable in a unique way (i.e., the
             given dimensions form a tree).

             A foreign key must have the same name as the primary key it
             references.

           - expectboguskeyvalues: If expectboguskeyvalues is True, we allow a
             key that is used as lookup attribute in a lower level to hold a
             wrong value (which would typically be None). When ensure or
             insert is called, we find the correct value for the key in the
             higher level.  If expectboguskeyvalues, we again try a lookup on
             the lower level after this. If expectboguskeyvalues is False, we
             move directly on to do an insert. Default: False
        """
        self.root = references[0][0]
        self.targetconnection = self.root.targetconnection
        self.key = self.root.key
        self.lookupatts = self.root.lookupatts

        dims = set([self.root])
        self.refs = {}
        self.refkeys = {}
        self.all = self.root.all[:]
        for (dim, refeddims) in references:
            # Check that all dimensions use the same target connection.
            # Build the dict self.refs: 
            #           {dimension -> set(refed dimensions)}
            # Build self.all from dimensions' lists
            # Keep track of seen dimensions by means of the set dims and
            # ensure that each table is only reachable once.
            if isinstance(refeddims, Dimension):
                # If there is only one dimension, then make a tuple with that
                refeddims = (refeddims, )
            for rd in refeddims:
                if rd.targetconnection is not self.targetconnection:
                    raise ValueError, "Different connections used"
                if rd in dims:
                    raise ValueError, "The tables do not form a tree"
                dims.add(rd)
                tmp = self.refs.get(dim, set())
                tmp.add(rd)
                self.refs[dim] = tmp
                # The key is alredy there as we assume FKs and PKs have
                # identical names
                self.all.extend(list(rd.attributes))

        # Check that all dimensions in dims are reachable from the root
        dimscopy = dims.copy()
        dimscopy.remove(self.root)
        for (tbl, targets) in self.refs.items():
            for target in targets:
                # It is safe to use remove as each dim is only referenced once
                dimscopy.remove(target)
        # Those dimensions that are left in dims at this point are unreachable
        if len(dimscopy) != 0:
            raise ValueError, "Not every given dimension is reachable"

        # Construct SQL...

        self.keylookupsql = self.root.keylookupsql

        self.allnames = []
        for dim in dims:
            for att in dim.attributes:
                self.allnames.append(att)

        # Make sure that there are no duplicated names:
        if len(self.allnames) != len(set(self.allnames)):
            raise ValueError, "Duplicated attribute names found"

        self.alljoinssql = "SELECT " + ", ".join(self.allnames) + \
            " FROM " + " NATURAL JOIN ".join(map(lambda d: d.name, dims))
        self.rowlookupsql = self.alljoinssql + " WHERE %s.%s = %%(%s)s" % \
            (self.root.name, self.root.key, self.root.key)

        self.levels = {}
        self.__buildlevels(self.root, 0)
        self.levellist = range(len(self.levels))
        self.levellist.reverse()

        self.expectboguskeyvalues = expectboguskeyvalues

    def __buildlevels(self, node, level):
        tmp = self.levels.get(level, [])
        tmp.append(node)
        self.levels[level] = tmp
        for ref in self.refs.get(node, []):
            self.__buildlevels(ref, level + 1)

    def lookup(self, row, namemapping={}):
        """ Find the key for the row with the given values.

            Arguments:
            - row: a dict which must contain at least the lookup attributes
              which all must come from the root (the table closest to the
              fact table).
            - namemapping: an optional namemapping (see module's documentation)
        """
        res = self._before_lookup(row, namemapping)
        if res:
            return res
        res = self.root.lookup(row, namemapping)
        self._after_lookup(row, namemapping, res)
        return res

    def _before_lookup(self, row, namemapping):
        return None

    def _after_lookup(self, row, namemapping, resultkeyvalue):
        pass

    def getbykey(self, keyvalue, fullrow=False):
        """Lookup and return the row with the given key value.

           If no row is found in the dimension table, the function returns
           a row where all values (including the key) are None.

           Arguments:
           - keyvalue: the key value of the row to lookup
           - fullrow: a flag deciding if the full row (with data from
             all tables in the snowflake) should be returned. If False,
             only data from the lowest level in the hierarchy (i.e., the table
             the closest to the fact table) is returned. Default: False
        """
        res = self._before_getbykey(keyvalue, fullrow)
        if res:
            return res
        if not fullrow:
            res = self.root.getbykey(keyvalue)
        else:
            self.targetconnection.execute(self.rowlookupsql, 
                                           {self.root.key : keyvalue})
            res = self.targetconnection.fetchone(self.allnames)
        self._after_getbykey(keyvalue, res, fullrow)
        return res

    def _before_getbykey(self, keyvalue, fullrow=False):
        return None

    def _after_getbykey(self, keyvalue, resultrow, fullrow=False):
        pass

    def getbyvals(self, values, namemapping={}, fullrow=False):
        """Return a list of all rows with values identical to the given.

           Arguments:
           - values: a dict which must hold a subset of the tables attributes.
             All rows that have identical values for all attributes in this
             dict are returned.
           - namemapping: an optional namemapping (see module's documentation)
           - fullrow: a flag deciding if the full row (with data from
             all tables in the snowflake) should be returned. If False,
             only data from the lowest level in the hierarchy (i.e., the table
             the closest to the fact table) is returned. Default: False
        """
        res = self._before_getbyvals(values, namemapping)
        if res is not None:
            return res

        if not fullrow:
            res = self.root.getbyvals(values, namemapping)
        else:
            # select all attributes from the table. 
            # The attributes available from the
            # values dict are used in the WHERE clause.
            attstouse = [a for a in self.allnames \
                             if a in values or a in namemapping] 
            sqlwhere = " WHERE " + \
                " AND ".join(["%s = %%(%s)s" % (att, att) for att in attstouse])
            self.targetconnection.execute(self.alljoinssql + sqlwhere, 
                                           values, namemapping)
            res = [r for r in self.targetconnection.rowfactory(self.allnames)]

        self._after_getbyvals(values, namemapping, res)
        return res

    def _before_getbyvals(self, values, namemapping, fullrow=False):
        return None

    def _after_getbyvals(self, values, namemapping, resultrows, fullrow=False):
        pass

    def update(self, row, namemapping={}):
        """Update rows in the participating dimension tables.

           If the key of a participating dimension D is in the given row, 
           D.update(...) is invoked.

           Note that this function is not good to use for updating a foreign
           key which here has the same name as the referenced primary key: The
           referenced table could then also get updated unless it is ensured
           that none of its attributes are present in the given row.

           In other words, it is often better to use the update function
           directly on the Dimensions that should be updated.
           
           Arguments:
            - row: a dict. If the key of a participating dimension D is in the
              dict, D.update(...) is invoked.
            - namemapping: an optional namemapping (see module's documentation)
        """
        res = self._before_update(row, namemapping)
        if res is not None:
            return

        for l in self.levellist:
            for t in self.levels[l]:
                if t.key in row or \
                        (t.key in namemapping and namemapping[t.key] in row):
                    t.update(row, namemapping)

        self._after_update(row, namemapping)

    def _before_update(self, row, namemapping):
        return None

    def _after_update(self, row, namemapping):
        pass

    def ensure(self, row, namemapping={}):
        """Lookup the given member. If that fails, insert it. Return key value.

           If the member must be inserted, data is automatically inserted in
           all participating tables where (part of) the member is not
           already represented.

           Key values for different levels may be added to the row. It is
           NOT guaranteed that key values for all levels exist in row
           afterwards.

           Arguments:
           - row: the row to lookup or insert. Must contain the lookup 
             attributes. 
           - namemapping: an optional namemapping (see module's documentation)
        """
        (key, ignored) = self.__ensure_helper(self.root, row, namemapping, 
                                              False)
        return key

    def insert(self, row, namemapping={}):
        """Insert the given member. If that fails, insert it. Return key value.

           Data is automatically inserted in all participating tables where
           (part of) the member is not already represented. If nothing is
           inserted at all, a ValueError is raised.

           Key values for different levels may be added to the row. It is
           NOT guaranteed that key values for all levels exist in row
           afterwards.

           Arguments:
           - row: the row to lookup or insert. Must contain the lookup 
             attributes. 
           - namemapping: an optional namemapping (see module's documentation)
        """
        key = self._before_insert(row, namemapping)
        if key is not None:
            return key
        (key, insertdone) = self.__ensure_helper(self.root, row, namemapping, 
                                                 False)
        if not insertdone:
            raise ValueError, "Member already present - nothing inserted"
        self._after_insert(row, namemapping, key)
        return key


    def _before_insert(self, row, namemapping):
        return None

    def _after_insert(self, row, namemapping, newkeyvalue):
        pass

    def endload(self):
        """Finalize the load."""
        pass

    def __ensure_helper(self, dimension, row, namemapping, insertdone):
        """ """
        # NB: Has side-effects: Key values are set for all dimensions
        key = None
        retry = False
        try:
            key = dimension.lookup(row, namemapping)
        except KeyError:
            retry = True # it can happen that the keys for the levels above 
                         # aren't there yet but should be used as lookup 
                         # attributes in dimension.
                         # Below we find them and we should then try a 
                         # lookup again before we move on to do an insertion
        if key is not None:
            row[(namemapping.get(dimension.key) or dimension.key)] = key
            return (key, insertdone)
        # Else recursively get keys for refed tables and then insert
        for refed in self.refs.get(dimension, []):
            (key, insertdone) = self.__ensure_helper(refed, row, namemapping, 
                                                     insertdone)
            # We don't need to set the key value in the row as this already 
            # happened in the recursive step.

        # We set insertdone = True to know later that we actually 
        #inserted something
        if retry or self.expectboguskeyvalues:
            # The following is similar to 
            #   key = dimension.ensure(row, namemapping)
            # but we set insertdone here.
            key = dimension.lookup(row, namemapping)
            if key is None:
                key = dimension.insert(row, namemapping)
                insertdone = True
        else:
            # We don't need to lookup again since no attributes were 
            # missing (no KeyError) and we don't expect bogus values. 
            # So we can proceed directly to do an insert.
            key = dimension.insert(row, namemapping)
            insertdone = True

        row[(namemapping.get(dimension.key) or dimension.key)] = key
        return (key, insertdone)

    def scdensure(self, row, namemapping={}):
        """Lookup or insert a version of a slowly changing dimension member.

           # Still experimental!!! For now we require that only the
           # root is a SlowlyChangingDimension.

           NB: Has side-effects on the given row.

           Arguments:
           - row: a dict containing the attributes for the member. 
           - namemapping: an optional namemapping (see module's documentation)
        """
        # Still experimental!!! For now we require that only the
        # root is a SlowlyChangingDimension.
        # If we were to allow other nodes to be SCDs, we should require
        # that those between those nodes and the root (incl.) were also
        # SCDs.
        for dim in self.levels.get(1, []):
            (keyval, ignored) = self.__ensure_helper(dim, row, namemapping, 
                                                     False)
            row[(namemapping.get(dim.key) or dim.key)] = keyval

        row[(namemapping.get(self.root.key) or self.root.key)] = \
            self.root.scdensure(row, namemapping)
        return row[(namemapping.get(self.root.key) or self.root.key)]


class FactTable(object):
    """A class for accessing a fact table in the DW."""

    def __init__(self, name, keyrefs, measures=(), targetconnection=None):
        """Arguments:
           - name: the name of the fact table in the DW
           - keyrefs: a sequence of attribute names that constitute the
             primary key of the fact tables (i.e., the dimension references)
           - measures: a possibly empty sequence of measure names. Default: ()
           - targetconnection: The ConnectionWrapper to use. If not given,
             the default target connection is used.
        """
        if targetconnection is None:
            targetconnection = pygrametl.getdefaulttargetconnection()
        self.targetconnection = targetconnection
        self.name = name
        self.keyrefs = keyrefs
        self.measures = measures
        self.all = [k for k in keyrefs] + [m for m in measures]
        pygrametl._alltables.append(self)

        # Create SQL

        # INSERT INTO name (key1, ..., keyn, meas1, ..., measn)
        # VALUES (%(key1)s, ..., %(keyn)s, %(meas1)s, ..., %(measn)s)
        self.insertsql = "INSERT INTO " + name + "(" + \
            ", ".join(keyrefs) + (measures and ", " or "") + \
            ", ".join(measures) + ") VALUES (" + \
            ", ".join(["%%(%s)s" % (att,) for att in self.all]) + ")"

        # SELECT key1, ..., keyn, meas1, ..., measn FROM name
        # WHERE key1 = %(key1)s AND ... keyn = %(keyn)s
        self.lookupsql = "SELECT " + ",".join(self.all) + " FROM " + name + \
            " WHERE " + " AND ".join(["%s = %%(%s)s" % (k, k) \
                                          for k in self.keyrefs])

    def insert(self, row, namemapping={}):
        """Insert a fact into the fact table.

           Arguments:
           - row: a dict at least containing values for the keys and measures.
           - namemapping: an optional namemapping (see module's documentation)
        """
        tmp = self._before_insert(row, namemapping)
        if tmp:
            return
        self.targetconnection.execute(self.insertsql, row, namemapping)
        self._after_insert(row, namemapping)

    def _before_insert(self, row, namemapping):
        return None

    def _after_insert(self, row, namemapping):
        pass

    def _emptyfacttonone(self, argdict):
        """Return None if the given argument only contains None values, 
           otherwise return the given argument
        """ 
        for k in self.keyrefs:
            if argdict[k] is not None:
                return argdict
        return None

    def lookup(self, keyvalues, namemapping={}):
        """Lookup a fact from the given key values. Return key and measure vals.

           Return None if no fact is found.

           Arguments:
           - keyvalues: a dict at least containing values for all keys
           - namemapping: an optional namemapping (see module's documentation)
        """
        res = self._before_lookup(keyvalues, namemapping)
        if res:
            return self._emptyfacttonone(res)
        self.targetconnection.execute(self.lookupsql, keyvalues, namemapping)
        res = self.targetconnection.fetchone(self.all)
        self._after_lookup(keyvalues, namemapping, res)
        return self._emptyfacttonone(res)

    def _before_lookup(self, keyvalues, namemapping):
        return None

    def _after_lookup(self, keyvalues, namemapping, resultrow):
        pass

    def ensure(self, row, compare=False, namemapping={}):
        """Ensure that a fact is present (insert it if it is not already there).

           Arguments:
           - row: a dict at least containing the attributes of the fact table
           - compare: a flag deciding if measure vales from a fact that was
             looked up are compared to those in the given row. If True and
             differences are found, a ValueError is raised. Default: False
           - namemapping: an optional namemapping (see module's documentation)
        """
        res = self.lookup(row, namemapping)
        if not res:
            self.insert(row, namemapping)
            return False
        elif compare:
            for m in self.measures:
                if m in row and row[m] != res.get(m):
                    raise ValueError, \
                        "The existing fact has different measure values"
        return True

    def endload(self):
        """Finalize the load."""
        pass


class BatchFactTable(FactTable):
    """A class for accessing a fact table in the DW. This class performs
       performs insertions in batches.
    """

    def __init__(self, name, keyrefs, measures=(), batchsize=10000,
                 targetconnection=None):
        """Arguments:
           - name: the name of the fact table in the DW
           - keyrefs: a sequence of attribute names that constitute the
             primary key of the fact tables (i.e., the dimension references)
           - measures: a possibly empty sequence of measure names. Default: ()
           - batchsize: an int deciding many insert operations should be done
             in one batch. Default: 10000
           - targetconnection: The ConnectionWrapper to use. If not given,
             the default target connection is used.
        """
        FactTable.__init__(self, name, keyrefs, measures, targetconnection)
        self.__batchsize = batchsize
        self.__batch = []

    def _before_insert(self, row, namemapping):
        self.__batch.append(pygrametl.project(self.all, row, namemapping))
        if len(self.__batch) == self.__batchsize:
            self.__insertnow()
        return True # signal that we did something

    def _before_lookup(self, keyvalues, namemapping):
        self.__insertnow()

    def endload(self):
        """Finalize the load."""
        self.__insertnow()

    def __insertnow(self):
        if self.__batch:
            self.targetconnection.executemany(self.insertsql, self.__batch)
            self.__batch = []


class BulkFactTable(object):
    """Class for addition of facts to a fact table. Reads are not supported. """

    def __init__(self, name, keyrefs, measures, bulkloader, 
                 fieldsep='\t', rowsep='\n', nullsubst=None,
                 tempdest=None, bulksize=500000, usefilename=False):
        """Arguments:
           - name: the name of the fact table in the DW
           - keyrefs: a sequence of attribute names that constitute the
             primary key of the fact tables (i.e., the dimension references)
           - measures: a possibly empty sequence of measure names.
           - bulkloader: A method 
             m(name, attributes, fieldsep, rowsep, nullsubst, tempdest) that
             is called to load data from a temporary file into the DW. The
             argument "attributes" is the combination of keyrefs and measures
             (i.e., a list of the names of the columns to insert values into)
             and show the order in which the attribute values appear in the
             temporary file.  The rest of the arguments are similar to those
             arguments with identical names that are given to
             BulkFactTable.__init__ as described here. The argument "tempdest"
             can, however, be 1) a string with a filename or 2) a file
             object. This is determined by the usefilename argument to
             BulkFactTable.__init__ (see below).
           - fieldsep: a string used to separate fields in the temporary 
             file. Default: '\\t'
           - rowsep: a string used to separate rows in the temporary file.
             Default: '\\n'
           - nullsubst: an optional string used to replace None values.
             If nullsubst=None, no substitution takes place. Default: None
           - tempdest: a file object or None. If None a named temporary file
             is used.
           - bulksize: an int deciding the number of rows to load in one
             bulk operation.
           - usefilename: a value deciding if the file should be passed to the
             bulkloader by its name instead of as a file-like object. This is
             necessary when the bulkloader runs in another process (for example,
             when if the BulkFactTable is wrapped by a DecoupledFactTable and
             invokes the bulkloader on a shared connection wrapper).
        """

        self.name = name
        self.keyrefs = keyrefs
        self.measures = measures
        self.all = [k for k in keyrefs] + [m for m in measures]
        self.__close = False
        if tempdest is None:
            self.__close = True
            self.__namedtempfile = tempfile.NamedTemporaryFile(bufsize=64*1024)
            tempdest = self.__namedtempfile.file
        self.fieldsep = fieldsep
        self.rowsep = rowsep
        self.nullsubst = nullsubst
        self.bulkloader = bulkloader
        self.tempdest = tempdest

        self.bulksize = bulksize
        self.usefilename = usefilename

        self.__count = 0
        self.__ready = True

        if nullsubst is None:
            self.insert = self._insertwithoutnulls
        else:
            self.insert = self._insertwithnulls

        pygrametl._alltables.append(self)

    def __preparetempfile(self):
        self.__namedtempfile = tempfile.NamedTemporaryFile()
        self.tempdest = self.__namedtempfile.file
        self.__ready = True

    def insert(self, row, namemapping={}):
        """Insert a fact into the fact table.

           Arguments:
           - row: a dict at least containing values for the keys and measures.
           - namemapping: an optional namemapping (see module's documentation)
        """
        pass # Is set to _insertwithnulls or _inserwithoutnulls from __init__

    def _insertwithnulls(self, row, namemapping={}):
        """Insert a fact into the fact table.

           Arguments:
           - row: a dict at least containing values for the keys and measures.
           - namemapping: an optional namemapping (see module's documentation)
        """
        if not self.__ready:
            self.__preparetempfile()
        rawdata = [row[namemapping.get(att) or att] for att in self.all]
        data = [pygrametl.getstrornullvalue(val, self.nullsubst) \
                for val in rawdata]
        self.__count += 1
        self.tempdest.write("%s%s" % (self.fieldsep.join(data), self.rowsep))
        if self.__count == self.bulksize:
            self.__bulkloadnow()

    def _insertwithoutnulls(self, row, namemapping={}):
        """Insert a fact into the fact table.

           Arguments:
           - row: a dict at least containing values for the keys and measures.
           - namemapping: an optional namemapping (see module's documentation)
        """
        if not self.__ready:
            self.__preparetempfile()
        data = [str(row[namemapping.get(att) or att]) for att in self.all]
        self.__count += 1
        self.tempdest.write("%s%s" % (self.fieldsep.join(data), self.rowsep))
        if self.__count == self.bulksize:
            self.__bulkloadnow()

    def __bulkloadnow(self):
        self.tempdest.flush()
        self.tempdest.seek(0)
        self.bulkloader(self.name, self.all, 
                        self.fieldsep, self.rowsep, self.nullsubst,
                        self.usefilename and self.__namedtempfile.name or \
                            self.tempdest)
        self.tempdest.seek(0)
        self.tempdest.truncate(0)
        self.__count = 0

    def endload(self):
        """Finalize the load."""
        if self.__count > 0:
            self.__bulkloadnow()
        if self.__close:
            try:
                self.__namedtempfile.close()
            except OSError:
                pass # may happen if the instance was decoupled
            self.__ready = False

    def _decoupled(self):
        if self.__close:
            # We need to make a private tempfile
            self.__namedtempfile = tempfile.NamedTemporaryFile()
            self.tempdest = self.__namedtempfile.file


class SubprocessFactTable(object):
    """Class for addition of facts to a subprocess. 

    The subprocess can, e.g., be a logger or bulkloader. Reads are not 
    supported. 

    Note that a created instance can not be used when endload() has been 
    called (and endload() is called from pygrametl.commit()).
    """

    def __init__(self, keyrefs, measures, executable, 
                 initcommand=None, endcommand=None, terminateafter=-1,
                 fieldsep='\t', rowsep='\n', nullsubst=None,
                 buffersize=16384):
        """Arguments:
           - keyrefs: a sequence of attribute names that constitute the
             primary key of the fact table (i.e., the dimension references)
           - measures: a possibly empty sequence of measure names. Default: ()
           - executable: The subprocess to start.
           - initcommand: If not None, this command is written to the
             subprocess before any data.
           - endcommand: If not None, this command is written to the subprocess
             after all data has been written.
           - terminateafter: If greater than or equal to 0, the subprocess
             is terminated after this amount of seconds after the pipe to
             the subprocess is closed.
           - fieldsep: a string used to separate fields in the output
             sent to the subprocess. Default: '\t'
           - rowsep: a string used to separate rows in the output sent to the
             subprocess. Default: '\n'
           - nullsubst: an optional string used to replace None values.
             If nullsubst=None, no substitution takes place. Default: None
        """

        self.all = [k for k in keyrefs] + [m for m in measures]
        self.keyrefs = keyrefs
        self.measures = measures
        self.endcommand = endcommand
        self.terminateafter = terminateafter
        self.fieldsep = fieldsep
        self.rowsep = rowsep
        self.nullsubst = nullsubst
        self.process = Popen(executable, bufsize=buffersize, shell=True, 
                             stdin=PIPE)
        self.pipe = self.process.stdin

        if nullsubst is None:
            self.insert = self._insertwithoutnulls
        else:
            self.insert = self._insertwithnulls

        if initcommand is not None:
            self.pipe.write(initcommand)

        pygrametl._alltables.append(self)

    def insert(self, row, namemapping={}):
        """Insert a fact into the fact table.

           Arguments:
           - row: a dict at least containing values for the keys and measures.
           - namemapping: an optional namemapping (see module's documentation)
        """
        pass # Is set to _insertwithnulls or _inserwithoutnulls from __init__

    def _insertwithnulls(self, row, namemapping={}):
        """Insert a fact into the fact table.

           Arguments:
           - row: a dict at least containing values for the keys and measures.
           - namemapping: an optional namemapping (see module's documentation)
        """
        rawdata = [row[namemapping.get(att) or att] for att in self.all]
        data = [pygrametl.getstrornullvalue(val, self.nullsubst) \
                for val in rawdata]
        self.pipe.write("%s%s" % (self.fieldsep.join(data), self.rowsep))

    def _insertwithoutnulls(self, row, namemapping={}):
        """Insert a fact into the fact table.

           Arguments:
           - row: a dict at least containing values for the keys and measures.
           - namemapping: an optional namemapping (see module's documentation)
        """
        data = [str(row[namemapping.get(att) or att]) for att in self.all]
        self.pipe.write("%s%s" % (self.fieldsep.join(data), self.rowsep))

    def endload(self):
        """Finalize the load."""
        if self.endcommand is not None:
            self.pipe.write(self.endcommand)
        
        self.pipe.close()
        if self.terminateafter >= 0:
            sleep(self.terminateafter)
            self.process.terminate()
        else:
            self.process.wait()

    def _decoupling(self):
        """Raise a TypeError to avoid decoupling (does not happen in Jython)"""
        import sys
        if sys.platform.startswith('java'):
            # In Jython, we use threads for decoupling and we do not have
            # to prevent it.
            return
        raise TypeError, 'A SubProcessFactTable cannot be decoupled'


class DecoupledDimension(pygrametl.parallel.Decoupled):
    """A Dimension-like class that enables parallelism by executing all 
       operations on a given Dimension in a separate, dedicated process 
       (that Dimension is said to be "decoupled").
    """

    def __init__(self, dim, returnvalues=True, consumes=(), attstoconsume=(),
                 batchsize=500, queuesize=200):
        """Arguments:
           - dim: the Dimension object to use in a separate process
           - returnvalues: decides if return values from method calls on dim
             should be kept such that they can be fetched by the caller or
             another Decoupled instance
           - consumes: a sequence of Decoupled objects from which to fetch
             returnvalues (that are used to replace FutureResults in arguments).
             Default: ()
           - attstoconsume: a sequence of the attribute names in rows that 
             should have FutureResults replaced by actual return values. Does
             not have to be given, but may improve performance when given.
             Default: ()
           - batchsize: the size of batches (grouped method calls) transferred
             between the processes. NB: Large values do not necessarily give
             good performance
             Default: 500
           - queuesize: the maximum amount of waiting batches. Infinite if
             less than or equal to 0. NB: Large values do not necessarily give
             good performance. 
             Default: 200
        """
        pygrametl.parallel.Decoupled.__init__(
            self, dim, returnvalues, consumes,
            tuple([(0, a) for a in attstoconsume]),
            batchsize, queuesize, False)
        if dim in pygrametl._alltables:
            pygrametl._alltables.remove(dim) # We add self instead...
        pygrametl._alltables.append(self)

    def lookup(self, row, namemapping={}):
        """Invoke lookup on the decoupled Dimension in the separate process"""
        return self._enqueue('lookup', row, namemapping)

    def getbykey(self, keyvalue):
        """Invoke getbykey on the decoupled Dimension in the separate process"""
        return self._enqueue('getbykey', keyvalue)

    def getbyvals(self, row, namemapping={}):
        "Invoke betbycals on the decoupled Dimension in the separate process"
        return self._enqueue('getbyvals', row, namemapping)

    def insert(self, row, namemapping={}):
        """Invoke insert on the decoupled Dimension in the separate process"""
        return self._enqueue('insert', row, namemapping)

    def ensure(self, row, namemapping={}):
        """Invoke ensure on the decoupled Dimension in the separate process"""
        return self._enqueue('ensure', row, namemapping)

    def endload(self):
        """Invoke endload on the decoupled Dimension in the separate process and
           return when all waiting method calls have been executed
        """
        self._enqueuenoreturn('endload') # first add 'endload' to the batch ...
        self._endbatch()                 # ... and then send the batch
        self._join()
        return None

    def scdensure(self, row, namemapping={}):
        "Invoke scdensure on the decoupled Dimension in the separate process"
        if hasattr(self._obj, 'scdensure'):
            return self._enqueue('scdensure', row, namemapping)
        else:
            raise AttributeError, 'The object does not support scdensure'


class DecoupledFactTable(pygrametl.parallel.Decoupled):
    """A FactTable-like class that enables parallelism by executing all 
       operations on a given FactTable in a separate, dedicated process 
       (that FactTable is said to be "decoupled").
    """

    def __init__(self, facttbl, returnvalues=True, consumes=(),
                 attstoconsume=(), batchsize=500, queuesize=200):
        """Arguments:
           - factbl: the FactTable object to use in a separate process
           - returnvalues: decides if return values from method calls on facttbl
             should be kept such that they can be fetched by the caller or
             another Decoupled instance
           - consumes: a sequence of Decoupled objects from which to fetch
             returnvalues (that are used to replace FutureResults in arguments).
             Default: ()
           - attstoconsume: a sequence of the attribute names in rows that 
             should have FutureResults replaced by actual return values. Does
             not have to be given, but may improve performance when given.
             Default: ()
           - batchsize: the size of batches (grouped method calls) transferred
             between the processes. NB: Large values do not necessarily give
             good performance
             Default: 500
           - queuesize: the maximum amount of waiting batches. Infinite if
             less than or equal to 0. NB: Large values do not necessarily give
             good performance. 
             Default: 200
        """
        pygrametl.parallel.Decoupled.__init__(
            self, facttbl, returnvalues, consumes,
            tuple([(0, a) for a in attstoconsume]),
            batchsize, queuesize, False)
        if facttbl in pygrametl._alltables:
            pygrametl._alltables.remove(facttbl) # We add self instead
        pygrametl._alltables.append(self)

    def insert(self, row, namemapping={}):
        """Invoke insert on the decoupled FactTable in the separate process"""
        return self._enqueue('insert', row, namemapping)

    def endload(self):
        """Invoke endload on the decoupled FactTable in the separate process and
           return when all waiting method calls have been executed
        """
        self._enqueuenoreturn('endload')
        self._endbatch()
        self._join()
        return None

    def lookup(self, row, namemapping={}):
        """Invoke lookup on the decoupled FactTable in the separate process"""
        if hasattr(self._obj, 'lookup'):
            return self._enqueue('lookup', row, namemapping)
        else:
            raise AttributeError, 'The object does not support lookup'
        
    def ensure(self, row, namemapping={}):
        """Invoke ensure on the decoupled FactTable in the separate process"""
        if hasattr(self._obj, 'ensure'):
            return self._enqueue('ensure', row, namemapping)
        else:
            raise AttributeError, 'The object does not support ensure'


class BasePartitioner(object):
    """A base class for partitioning between several parts. 

       See also DimensionPartitioner and FactTablePartitioner.
    """
    def __init__(self, parts):
        self.parts = list(parts)
        self.__nextpart = 0

    def parts(self):
        """Return the parts the partitioner works on"""
        return self.parts[:]

    def addpart(self, part):
        """Add a part"""
        self.parts.append(part)

    def droppart(self, part=None):
        """Drop a part. If an argument is given, it must be a part of the
           patitioner and it will then be removed. If no argument is given,
           the first part is removed."""
        if part is None:
            self.parts.pop()
        else:
            self.parts.remove(part)

    def getpart(self, row, namemapping={}):
        """Find  the part that should handle the given row. The provided
        implementation in BasePartitioner does only use round robin 
        partitioning, but subclasses apply other methods """
        part = self.parts[self.__nextpart]
        self.__nextpart = (self.__nextpart + 1) % len(self.parts)
        return part

    def endload(self):
        """Call endload on all parts"""
        for part in self.parts:
            part.endload()


class DimensionPartitioner(BasePartitioner):
    """A Dimension-like class that handles partitioning.

       Partitioning is done between a number of Dimension objects called the
       parts. The class offers the interface of Dimensions (incl. scdensure
       from SlowlyChangingDimension). When a method is called, the
       corresponding method on one of the parts (chosen by a user-definable
       partitioner function) will be invoked. The parts can operate on a
       single physical dimension table or different physical tables.
    """

    def __init__(self, parts, getbyvalsfromall=False, partitioner=None):
        """
        Arguments:
        - parts: a sequence of Dimension objects.
        - getbyvalsfromall: determines if getbyvals should be answered by means
          of all parts (when getbyvalsfromall = True) or only the first part, 
          i.e., parts[0] (when getbybalsfromall = False). Default: False
        - partitioner: None or a callable p(dict) -> int where the argument
          is a dict mapping from the names of the lookupatts to the values of
          the lookupatts. The resulting int is used to determine which part
          a given row should be handled by.
          When partitioner is None, a default partitioner is used. This
          partitioner computes the hash value of each value of the lookupatts
          and adds them together.
        """
        BasePartitioner.__init__(self, parts)
        self.getbyvalsfromall = getbyvalsfromall
        self.lookupatts = parts[0].lookupatts
        self.key = parts[0].key
        for p in parts:
            if not p.lookupatts == self.lookupatts:
                raise ValueError, 'The parts must have the same lookupatts'
            if not p.key == self.key:
                raise ValueError, 'The parts must have the same key'
        if partitioner is not None:
            self.partitioner = partitioner
        else:
            # A partitioner that takes the hash of each attribute value in 
            # row and adds them all together:
            # Reading from right to left: get the values, use hash() on each
            # of them, and add all the hash values
            self.partitioner = lambda row: reduce((lambda x,y: x+y),\
                                                      map(hash, row.values()))

    def getpart(self, row, namemapping={}):
        """Return the part that should handle the given row"""
        vals = {}
        for att in self.lookupatts:
            vals[att] = row[namemapping.get(att) or att]
        return self.parts[self.partitioner(vals) % len(self.parts)]
    
    # Below this, methods like those in Dimensions:

    def lookup(self, row, namemapping={}):
        """Invoke lookup on the relevant Dimension part"""
        part = self.getpart(row, namemapping)
        return part.lookup(row, namemapping)

    def __getbykeyhelper(self, keyvalue):
        # Returns (rowresult, part). part is None if no result was found.
        for part in self.parts:
            row = part.getbykey(keyvalue)
            if row[self.key] is not None:
                return (row, part)
        return (row, None)

    def getbykey(self, keyvalue):
        """Invoke getbykey on the relevant Dimension part"""
        return self.__getbykeyhelper(keyvalue)[0]

    def getbyvals(self, values, namemapping={}):
        """Invoke getbyvals on the first part or all parts (depending on the
           value of the instance's getbyvalsfromall)"""
        if not self.getbyvalsfromall:
            return self.parts[0].getbyvals(values, namemapping)
        res = []
        for part in self.parts:
            res += part.getbyvals(values, namemapping)
        return res
        
    def update(self, row, namemapping={}):
        """Invoke update on the relevant Dimension part"""
        keyval = row[namemapping.get(self.key) or self.key]
        part = self.__getbykeyhelper(keyval)[1]
        if part is not None:
            part.update(row, namemapping)

    def ensure(self, row, namemapping={}):
        """Invoke ensure on the relevant Dimension part"""
        part = self.getpart(row, namemapping)
        return part.ensure(row, namemapping)

    def insert(self, row, namemapping={}):
        """Invoke insert on the relevant Dimension part"""
        part = self.getpart(row, namemapping)
        return part.insert(row, namemapping)

    def scdensure(self, row, namemapping={}):
        """Invoke scdensure on the relevant Dimension part"""
        part = self.getpart(row, namemapping)
        return part.scdensure(row, namemapping)


class FactTablePartitioner(BasePartitioner):
    """A FactTable-like class that handles partitioning.

       Partitioning is done between a number of FactTable objects called the
       parts. The class offers the interface of FactTable. When a method is
       called, the corresponding method on one of the parts (chosen by a
       user-definable partitioner function) will be invoked. The parts can
       operate on a single physical fact table or different physical
       tables. """

    def __init__(self, parts, partitioner=None):
        """
        Arguments:
        - parts: a sequence of FactTable objects.
        - partitioner: None or a callable p(dict) -> int where the argument
          is a dict mapping from the names of the keyrefs to the values of
          the keyrefs. The resulting int is used to determine which part
          a given row should be handled by.
          When partitioner is None, a default partitioner is used. This
          partitioner computes the sum of all the keyrefs values.
        """
        BasePartitioner.__init__(self, parts)
        if partitioner is not None:
            self.partitioner = partitioner
        else:
            self.partitioner = lambda row: reduce((lambda x, y: x+y), \
                                                      row.values())
        self.all = parts[0].all
        self.keyrefs = parts[0].keyrefs
        self.measures = parts[0].measures
        for ft in parts:
            if not (self.keyrefs == ft.keyrefs and \
                        self.measures == ft.measures):
                raise ValueError, \
                    'The parts must have the same measures and keyrefs'

    def getpart(self, row, namemapping={}):
        """Return the relevant part for the given row """
        vals = {}
        for att in self.keyrefs:
            vals[att] = row[namemapping.get(att) or att]
        return self.parts[self.partitioner(vals) % len(self.parts)]

    def insert(self, row, namemapping={}):
        """Invoke insert on the relevant part """
        part = self.getpart(row, namemapping)
        part.insert(row, namemapping)

    def lookup(self, row, namemapping={}):
        """Invoke lookup on the relevant part """
        part = self.getpart(row, namemapping)
        return part.lookup(row, namemapping)        

    def ensure(self, row, namemapping={}):
        """Invoke ensure on the relevant part """
        part = self.getpart(row, namemapping)
        return part.ensure(row, namemapping)        





# -*- coding: utf-8 -*-
"""
   A package for creating Extract-Transform-Load (ETL) programs in Python.

   The package contains a number of classes for filling fact tables
   and dimensions (including snowflaked and slowly changing dimensions), 
   classes for extracting data from different sources, classes for defining
   'steps' in an ETL flow, and convenient functions for often-needed ETL
   functionality.

   The package's modules are:
   - datasources for access to different data sources
   - tables for giving easy and abstracted access to dimension and fact tables
   - parallel for parallelizing ETL operations
   - JDBCConnectionWrapper and jythonmultiprocessing for support of Jython
   - aggregators for aggregating data
   - steps for defining steps in an ETL flow
   - FIFODict for providing a dict with a limited size and where elements are 
     removed in first-in first-out order
"""

import copy as pcopy
import exceptions
import types
from datetime import date, datetime
from Queue import Queue
from sys import modules
from threading import Thread

import FIFODict

__author__ = "Christian Thomsen"
__maintainer__ = "Sílex Sistemas Ltda."
__version__ = '0.0.1'

__all__ = ['project', 'copy', 'renamefromto', 'rename', 'renametofrom', 
           'getint', 'getlong', 'getfloat', 'getstr', 'getstrippedstr', 
           'getstrornullvalue', 'getbool', 'getdate', 'gettimestamp', 
           'getvalue', 'getvalueor', 'setdefaults', 'rowfactory', 'endload', 
           'today', 'now', 'ymdparser', 'ymdhmsparser', 'datereader', 
           'datetimereader', 'datespan', 'toupper', 'tolower', 'keepasis', 
           'ConnectionWrapper']


_alltables = []


def project(atts, row, renaming={}):
    """Create a new dictionary with a subset of the attributes.

       Arguments:
       - atts is a sequence of attributes in row that should be copied to the
         new result row.
       - row is the original dictionary to copy data from.
       - renaming is a mapping of names such that for each k in atts, 
         the following holds:
         - If k in renaming then result[k] = row[renaming[k]]. 
         - If k not in renaming then result[k] = row[k].
         renaming defauts to {}
         
    """
    res = {}
    for c in atts:
        if c in renaming:
            res[c] = row[renaming[c]]
        else:
            res[c] = row[c]
    return res


def copy(row, **renaming):
    """Create a copy of a dictionary, but allow renamings.

       Arguments:
       - row the dictionary to copy
       - **renaming allows renamings to be specified in the form
         newname=oldname meaning that in the result oldname will be
         renamed to newname.
    """
    if not renaming:
        return row.copy()

    tmp = row.copy()
    res = {}
    for k,v in renaming.items():
        res[k] = row[v]
        del tmp[v]
    res.update(tmp)
    return res


def renamefromto(row, renaming):
    """Rename keys in a dictionary.

       For each (oldname, newname) in renaming.items(): rename row[oldname] to 
       row[newname].
    """
    if not renaming:
        return row

    for old, new in renaming.items():
        row[new] = row[old]
        del row[old]

rename = renamefromto # for backwards compatibility


def renametofrom(row, renaming):    
    """Rename keys in a dictionary.

       For each (newname, oldname) in renaming.items(): rename row[oldname] to 
       row[newname].
    """
    if not renaming:
        return row

    for new, old in renaming.items():
        row[new] = row[old]
        del row[old]

    
def getint(value, default=None):
    """getint(value[, default]) -> int(value) if possible, else default."""
    try:
        return int(value)
    except Exception:
        return default


def getlong(value, default=None):
    """getlong(value[, default]) -> long(value) if possible, else default."""
    try:
        return long(value)
    except Exception:
        return default


def getfloat(value, default=None):
    """getfloat(value[, default]) -> float(value) if possible, else default."""
    try:
        return float(value)
    except Exception:
        return default


def getstr(value, default=None):
    """getstr(value[, default]) -> str(value) if possible, else default."""
    try:
        return str(value)
    except Exception:
        return default


def getstrippedstr(value, default=None):
    """Convert given value to a string and use .strip() on the result.

       If the conversion fails, the given default value is returned.
    """
    try:
        s = str(value)
        return s.strip()
    except Exception:
        return default


def getstrornullvalue(value, nullvalue='None'):
    """Convert a given value different from None to a string.

       If the given value is None, nullvalue (default: 'None') is returned.
    """
    if value is None:
        return nullvalue
    else:
        return str(value)


def getbool(value, default=None, 
            truevalues=set(( True,  1, '1', 't', 'true',  'True' )), 
            falsevalues=set((False, 0, '0', 'f', 'false', 'False'))):
    """Convert a given value to True, False, or a default value.

       If the given value is in the given truevalues, True is returned.
       If the given value is in the given falsevalues, False is returned.
       Otherwise, the default value is returned.
    """
    if value in truevalues:
        return True
    elif value in falsevalues:
        return False
    else:
        return default


def getdate(targetconnection, ymdstr, default=None):
    """Convert a string of the form 'yyyy-MM-dd' to a Date object.

       The returned Date is in the given targetconnection's format.
       Arguments:
       - targetconnection: a ConnectionWrapper whose underlying module's
         Date format is used
       - ymdstr: the string to convert
       - default: The value to return if the conversion fails
    """
    try:
        (year, month, day) = ymdstr.split('-')
        modref = targetconnection.getunderlyingmodule()
        return modref.Date(int(year), int(month), int(day))
    except Exception:
        return default


def gettimestamp(targetconnection, ymdhmsstr, default=None):
    """Converts a string of the form 'yyyy-MM-dd HH:mm:ss' to a Timestamp.
    
    The returned Timestamp is in the given targetconnection's format.
       Arguments:
       - targetconnection: a ConnectionWrapper whose underlying module's
         Timestamp format is used
       - ymdhmsstr: the string to convert
       - default: The value to return if the conversion fails
    """
    try:
        (datepart, timepart) = ymdhmsstr.strip().split(' ')
        (year, month, day) = datepart.split('-')
        (hour, minute, second) = timepart.split(':')
        modref = targetconnection.getunderlyingmodule()
        return modref.Timestamp(int(year), int(month), int(day),\
                                int(hour), int(minute), int(second))
    except Exception:
        return default


def getvalue(row, name, mapping={}):
    """If name in mapping, return row[mapping[name]], else return row[name]."""
    if name in mapping:
        return row[mapping[name]]
    else:
        return row[name]


def getvalueor(row, name, mapping={}, default=None):
    """Return the value of name from row using a mapping and a default value."""
    if name in mapping:
        return row.get(mapping[name], default)
    else:
        return row.get(name, default)


def setdefaults(row, attributes, defaults=None):
    """Set default values for attributes not present in a dictionary.

       Default values are set for "missing" values, existing values are not 
       updated. 

       Arguments:
       - row is the dictionary to set default values in
       - attributes is either
           A) a sequence of attribute names in which case defaults must
              be an equally long sequence of these attributes default values or
           B) a sequence of pairs of the form (attribute, defaultvalue) in 
              which case the defaults argument should be None
       - defaults is a sequence of default values (see above)
    """
    if defaults and len(defaults) != len(attributes):
        raise ValueError, "Lists differ in length"

    if defaults:
        seqlist = zip(attributes, defaults)
    else:
        seqlist = attributes

    for att, defval in seqlist:
        if att not in row:
            row[att] = defval


def rowfactory(source, names, close=True):
    """Generate dicts with key values from names and data values from source.
    
       The given source should provide either next() or fetchone() returning
       a tuple or fetchall() returning a sequence of tuples. For each tuple,
       a dict is constructed such that the i'th element in names maps to 
       the i'th value in the tuple.

       If close=True (the default), close will be called on source after
       fetching all tuples.
    """
    nextfunc = getattr(source, 'next', None)
    if nextfunc is None:
        nextfunc = getattr(source, 'fetchone', None)

    try:
        if nextfunc is not None:
            try:
                tmp = nextfunc()
                if tmp is None:
                    return
                else:
                    yield dict(zip(names, tmp))
            except (StopIteration, IndexError):
                return
        else:
            for row in source.fetchall():
                yield dict(zip(names, row))
    finally:
        if close:
            try:
                source.close()
            except:
                return


def endload():
    """Signal to all Dimension and FactTable objects that all data is loaded."""
    global _alltables
    for t in _alltables:
        method = getattr(t, 'endload', None)
        if callable(method):
            method()

_today = None
def today(ignoredtargetconn=None, ignoredrow=None, ignorednamemapping=None):
    """Return the date of the first call this method as a datetime.date object.
    """
    global _today
    if _today is not None:
        return _today
    _today = date.today()
    return _today

_now = None
def now(ignoredtargetconn=None, ignoredrow=None, ignorednamemapping=None):
    """Return the time of the first call this method as a datetime.datetime.
    """
    global _now
    if _now is not None:
        return _now
    _now = datetime.now()
    return _now


def ymdparser(ymdstr):
    """Convert a string of the form 'yyyy-MM-dd' to a datetime.date. 

       If the input is None, the return value is also None.
    """
    if ymdstr is None:
        return None
    (year, month, day) = ymdstr.split('-')
    return date(int(year), int(month), int(day))


def ymdhmsparser(ymdhmsstr):
    """Convert a string 'yyyy-MM-dd HH:mm:ss' to a datetime.datetime. 

       If the input is None, the return value is also None.
    """
    if ymdhmsstr is None:
        return None
    (datepart, timepart) = ymdhmsstr.strip().split(' ')
    (year, month, day) = datepart.split('-')
    (hour, minute, second) = timepart.split(':')
    return datetime(int(year), int(month), int(day),\
                    int(hour), int(minute), int(second))


def datereader(dateattribute, parsingfunction=ymdparser):
    """Return a function that converts a certain dict member to a datetime.date

       When setting, fromfinder for a tables.SlowlyChangingDimension, this
       method can be used for generating a function that picks the relevant
       dictionary member from each row and converts it.

       Arguments:
       - dateattribute: the attribute the generated function should read
       - parsingfunction: the parsing function that converts the string
         to a datetime.date
    """
    def readerfunction(targetconnection, row, namemapping = {}):
        atttouse = (namemapping.get(dateattribute) or dateattribute)
        return parsingfunction(row[atttouse]) # a datetime.date
    
    return readerfunction
    

def datetimereader(datetimeattribute, parsingfunction=ymdhmsparser):
    """Return a function that converts a certain dict member to a datetime

       When setting, fromfinder for a tables.SlowlyChangingDimension, this
       method can be used for generating a function that picks the relevant
       dictionary member from each row and converts it.

       Arguments:
       - datetimeattribute: the attribute the generated function should read
       - parsingfunction: the parsing function that converts the string
         to a datetime.datetime
    """
    def readerfunction(targetconnection, row, namemapping = {}):
        atttouse = (namemapping.get(datetimeattribute) or datetimeattribute)
        return parsingfunction(row[atttouse]) # a datetime.datetime

    return readerfunction


def datespan(fromdate, todate, fromdateincl=True, todateincl=True,
             key='dateid', 
             strings={'date':'%Y-%m-%d', 'monthname':'%B', 'weekday':'%A'},
             ints={'year':'%Y', 'month':'%m', 'day':'%d'},
             expander=None):
    """ Return a generator yielding dicts for all dates in an interval.

        Arguments:
        - fromdate: The lower bound for the date interval. Should be a
          datetime.date or a YYYY-MM-DD formatted string.
        - todate: The upper bound for the date interval. Should be a
          datetime.date or a YYYY-MM-DD formatted string.
        - fromdateincl: Decides if fromdate is included. Default: True
        - todateincl: Decides if todate is included. Default: True
        - key: The name of the attribute where an int (YYYYMMDD) that uniquely
          identifies the date is stored. Default: 'dateid'.
        - strings: A dict mapping attribute names to formatting directives (as
          those used by strftime). The returned dicts will have the specified
          attributes as strings.
          Default: {'date':'%Y-%m-%d', 'monthname':'%B', 'weekday':'%A'}
        - ints: A dict mapping attribute names to formatting directives (as
          those used by strftime). The returned dicts will have the specified
          attributes as ints.
          Default: {'year':'%Y', 'month':'%m', 'day':'%d'}
        - expander: A callable f(date, dict) that is invoked on each created
          dict. Not invoked if None. Default: None
    """

    for arg in (fromdate, todate):
        if not ((type(arg) in types.StringTypes and arg.count('-') == 2)\
                    or isinstance(arg, date)):
            raise ValueError, \
            "fromdate and today must be datetime.dates or " + \
            "YYYY-MM-DD formatted strings"

    (year, month, day) = fromdate.split('-')
    fromdate = date(int(year), int(month), int(day))

    (year, month, day) = todate.split('-')
    todate = date(int(year), int(month), int(day))

    start = fromdate.toordinal()
    if not fromdateincl:
        start += 1

    end = todate.toordinal()
    if todateincl:
        end += 1

    for i in xrange(start, end):
        d = date.fromordinal(i)
        res = {}
        res[key] = int(d.strftime('%Y%m%d'))
        for (att, format) in strings.iteritems():
            res[att] = d.strftime(format)
        for (att, format) in ints.iteritems():
            res[att] = int(d.strftime(format))
        if expander is not None:
            expander(d, res)
        yield res


toupper  = lambda s: s.upper()
tolower  = lambda s: s.lower()
keepasis = lambda s: s

_defaulttargetconnection = None

def getdefaulttargetconnection():
    """Return the default target connection"""
    global _defaulttargetconnection
    return _defaulttargetconnection


class ConnectionWrapper(object):
    """Provide a uniform representation of different database connection types.

       All Dimensions and FactTables communicate with the data warehouse using 
       a ConnectionWrapper. In this way, the code for loading the DW does not
       have to care about which parameter format is used.  
       
       pygrametl's code uses the 'pyformat' but the ConnectionWrapper performs
       translations of the SQL to use 'named', 'qmark', 'format', or 'numeric'
       if the user's database connection needs this. Note that the
       translations are simple and naive. Escaping as in %%(name)s is not
       taken into consideration. These simple translations are enough for
       pygrametl's code which is the important thing here; we're not trying to
       make a generic, all-purpose tool to get rid of the problems with
       different parameter formats. It is, however, possible to disable the
       translation of a statement to execute such that 'problematic'
       statements can be executed anyway.
    """

    def __init__(self, connection, stmtcachesize=1000):
        """Create a ConnectionWrapper around the given PEP 249 connection 

           If no default ConnectionWrapper already exists, the new 
           ConnectionWrapper is set as the default.

           Arguments:
           - connection: An open PEP 249 connection to the database
           - stmtcachesize: A number deciding how many translated statements to
             cache. A statement needs to be translated when the connection
             does not use 'pyformat' to specify parameters. When 'pyformat' is
             used, stmtcachesize is ignored as no statements need to be 
             translated. 
        """
        self.__connection = connection
        self.__cursor = connection.cursor()
        self.nametranslator = lambda s: s
        try:
            paramstyle = \
                modules[self.__connection.__class__.__module__].paramstyle
        except AttributeError:
            # Note: This is probably a better way to do it, but to avoid to
            # break anything that worked before this fix, we only do it this 
            # way if the first approach didn't work
            paramstyle = \
                modules[self.__connection.__class__.__module__.split('.')[0]].\
                paramstyle
        if not paramstyle == 'pyformat':
            self.__translations = FIFODict.FIFODict(stmtcachesize)
            try:
                self.__translate = getattr(self, '_translate2' + paramstyle)
            except AttributeError:
                raise InterfaceError, "The paramstyle '%s' is not supported" %\
                    paramstyle
        else:
            self.__translate = None

        global _defaulttargetconnection
        if _defaulttargetconnection is None:
            _defaulttargetconnection = self

    def execute(self, stmt, arguments=None, namemapping=None, translate=True):
        """Execute a statement.

           Arguments:
           - stmt: the statement to execute
           - arguments: a mapping with the arguments (default: None)
           - namemapping: a mapping of names such that if stmt uses %(arg)s
             and namemapping[arg]=arg2, the value arguments[arg2] is used 
             instead of arguments[arg]
           - translate: decides if translation from 'pyformat' to the
             undlying connection's format should take place. Default: True
        """
        if namemapping and arguments:
            arguments = copy(arguments, **namemapping)
        if self.__translate and translate:
            (stmt, arguments) = self.__translate(stmt, arguments)
        self.__cursor.execute(stmt, arguments)

    def executemany(self, stmt, params, translate=True):
        """Execute a sequence of statements."""
        if self.__translate and translate:
            # Idea: Translate the statement for the first parameter set. Then
            # reuse the statement (but create new attribute sequences if needed)
            # for the remaining paramter sets
            newstmt = self.__translate(stmt, params[0])[0]
            if type(self.__translations[stmt]) == str:
                # The paramstyle is 'named' in this case and we don't have to
                # put parameters into sequences
                self.__cursor.executemany(newstmt, params)
            else:
                # We need to extract attributes and put them into sequences
                names = self.__translations[stmt][1] # The attributes to extract
                newparams = [[p[n] for n in names] for p in params]
                self.__cursor.executemany(newstmt, newparams)
        else:
            # for pyformat when no translation is necessary
            self.__cursor.executemany(stmt, params)

    def _translate2named(self, stmt, row=None):
        # Translate %(name)s to :name. No need to change row.
        # Cache only the translated SQL.
        res = self.__translations.get(stmt, None)
        if res:
            return (res, row)
        res = stmt
        while True:
            start = res.find('%(')
            if start == -1: 
                break
            end = res.find(')s', start)
            if end == -1: 
                break
            name = res[start+2 : end]
            res = res.replace(res[start:end+2], ':' + name)
        self.__translations[stmt] = res
        return (res, row)

    def _translate2qmark(self, stmt, row=None):
        # Translate %(name)s to ? and build a list of attributes to extract
        # from row. Cache both.
        (newstmt, names) = self.__translations.get(stmt, (None, None))
        if newstmt:
            return (newstmt, [row[n] for n in names])
        names = []
        newstmt = stmt
        while True:
            start = newstmt.find('%(')
            if start == -1:
                break
            end = newstmt.find(')s', start)
            if end == -1: 
                break
            name = newstmt[start+2 : end]
            names.append(name)
            newstmt = newstmt.replace(newstmt[start:end+2], '?',1)#Replace once!
        self.__translations[stmt] = (newstmt, names)
        return (newstmt, [row[n] for n in names])

    def _translate2numeric(self, stmt, row=None):
        # Translate %(name)s to 1,2,... and build a list of attributes to
        # extract from row. Cache both.
        (newstmt, names) = self.__translations.get(stmt, (None, None))
        if newstmt:
            return (newstmt, [row[n] for n in names])
        names = []
        cnt = 0
        newstmt = stmt
        while True:
            start = newstmt.find('%(')
            if start == -1:
                break
            end = newstmt.find(')s', start)
            if end == -1: 
                break
            name = newstmt[start+2 : end]
            names.append(name)
            newstmt = newstmt.replace(newstmt[start:end+2], ':' + str(cnt))
            cnt += 1
        self.__translations[stmt] = (newstmt, names)
        return (newstmt, [row[n] for n in names])

    def _translate2format(self, stmt, row=None):
        # Translate %(name)s to %s and build a list of attributes to extract
        # from row. Cache both.
        (newstmt, names) = self.__translations.get(stmt, (None, None))
        if newstmt:
            return (newstmt, [row[n] for n in names])
        names = []
        newstmt = stmt
        while True:
            start = newstmt.find('%(')
            if start == -1:
                break
            end = newstmt.find(')s', start)
            if end == -1: 
                break
            name = newstmt[start+2 : end]
            names.append(name)
            newstmt = newstmt.replace(newstmt[start:end+2],'%s',1)#Replace once!
        self.__translations[stmt] = (newstmt, names)
        return (newstmt, [row[n] for n in names])

    def rowfactory(self, names=None):
        """Return a generator object returning result rows (i.e. dicts)."""
        rows = self.__cursor
        self.__cursor = self.__connection.cursor()
        if names is None:
            if rows.description is None: # no query was executed ...
                return (nothing for nothing in []) # a generator with no rows
            else:
                names = [self.nametranslator(t[0]) for t in rows.description]
        return rowfactory(rows, names, True)

    def fetchone(self, names=None):
        """Return one result row (i.e. dict)."""
        if self.__cursor.description is None:
            return {}
        if names is None:
            names = [self.nametranslator(t[0]) \
                         for t in self.__cursor.description]
        values = self.__cursor.fetchone()
        if values is None:
            return dict([(n, None) for n in names])#A row with each att = None
        else:
            return dict(zip(names, values))

    def fetchonetuple(self):
        """Return one result tuple."""
        if self.__cursor.description is None:
            return ()
        values = self.__cursor.fetchone()
        if values is None:
            return (None, ) * len(self.__cursor.description)
        else:
            return values

    def fetchmanytuples(self, cnt):
        """Return cnt result tuples."""
        if self.__cursor.description is None:
            return []
        return self.__cursor.fetchmany(cnt)

    def fetchalltuples(self):
        """Return all result tuples"""
        if self.__cursor.description is None:
            return []
        return self.__cursor.fetchall()

    def rowcount(self):
        """Return the size of the result."""
        return self.__cursor.rowcount

    def getunderlyingmodule(self):
        """Return a reference to the underlying connection's module."""
        return modules[self.__connection.__class__.__module__]

    def commit(self):
        """Commit the transaction."""
        endload()
        self.__connection.commit()

    def close(self):
        """Close the connection to the database,"""
        self.__connection.close()

    def rollback(self):
        """Rollback the transaction."""
        self.__connection.rollback()

    def setasdefault(self):
        """Set this ConnectionWrapper as the default connection."""
        global _defaulttargetconnection
        _defaulttargetconnection = self

    def cursor(self):
        """Return a cursor object. Optional method."""
        return self.__connection.cursor()

    def resultnames(self):
        if self.__cursor.description is None:
            return None
        else:
            return tuple([t[0] for t in self.__cursor.description])

    def __getstate__(self):
        # In case the ConnectionWrapper is pickled (to be sent to another
        # process), we need to create a new cursor when it is unpickled.
        res = self.__dict__.copy()
        del res['_ConnectionWrapper__cursor'] # a dirty trick, but...
        return res

    def __setstate__(self, dict):
        self.__dict__.update(dict)
        self.__cursor = self.__connection.cursor()


class BackgroundConnectionWrapper(object):
    """An alternative implementation of the ConnectionWrapper for experiments.
       This implementation communicates with the database by using a
       separate thread.

       It is likely better to use ConnectionWrapper og a shared 
       ConnectionWrapper (see pygrametl.parallel).

       This class offers the same methods as ConnectionWrapper. The 
       documentation is not repeated here.
    """
    _SINGLE = 1
    _MANY = 2

    # Most of this class' code was just copied from ConnectionWrapper
    # as we just want to do experiments with this class.

    def __init__(self, connection, stmtcachesize=1000):
        self.__connection = connection
        self.__cursor = connection.cursor()
        self.nametranslator = lambda s: s
        try:
            paramstyle = \
                modules[self.__connection.__class__.__module__].paramstyle
        except AttributeError:
            # Note: This is probably a better way to do it, but to avoid to
            # break anything that worked before this fix, we only do it this 
            # way if the first approach didn't work
            paramstyle = \
                modules[self.__connection.__class__.__module__.split('.')[0]].\
                paramstyle
        if not paramstyle == 'pyformat':
            self.__translations = FIFODict.FIFODict(stmtcachesize)
            try:
                self.__translate = getattr(self, '_translate2' + paramstyle)
            except AttributeError:
                raise InterfaceError, "The paramstyle '%s' is not supported" %\
                    paramstyle
        else:
            self.__translate = None

        # Thread-stuff
        self.__cursor = connection.cursor()
        self.__queue = Queue(5000)
        t = Thread(target=self.__worker)
        t.daemon = True
        t.start()

    def execute(self, stmt, arguments=None, namemapping=None, translate=True):
        if namemapping and arguments:
            arguments = copy(arguments, **namemapping)
        if self.__translate and translate:
            (stmt, arguments) = self.__translate(stmt, arguments)
        self.__queue.put((self._SINGLE, self.__cursor, stmt, arguments)) 

    def executemany(self, stmt, params, translate=True):
        if self.__translate and translate:
            # Idea: Translate the statement for the first parameter set. Then
            # reuse the statement (but create new attribute sequences if needed)
            # for the remaining paramter sets
            newstmt = self.__translate(stmt, params[0])[0]
            if type(self.__translations[stmt]) == str:
                # The paramstyle is 'named' in this case and we don't have to
                # put parameters into sequences
                self.__queue.put((self._MANY, self.__cursor, newstmt, params))
            else:
                # We need to extract attributes and put them into sequences
                names = self.__translations[stmt][1] # The attributes to extract
                newparams = [[p[n] for n in names] for p in params]
                self.__queue.put((self._MANY,self.__cursor, newstmt, newparams))
        else:
            # for pyformat when no translation is necessary
            self.__queue.put((self._MANY, self.__cursor, stmt, params))

    def _translate2named(self, stmt, row=None):
        # Translate %(name)s to :name. No need to change row.
        # Cache only the translated SQL.
        res = self.__translations.get(stmt, None)
        if res:
            return (res, row)
        res = stmt
        while True:
            start = res.find('%(')
            if start == -1:
                break
            end = res.find(')s', start)
            name = res[start+2 : end]
            res = res.replace(res[start:end+2], ':' + name)
        self.__translations[stmt] = res
        return (res, row)

    def _translate2qmark(self, stmt, row=None):
        # Translate %(name)s to ? and build a list of attributes to extract
        # from row. Cache both.
        (newstmt, names) = self.__translations.get(stmt, (None, None))
        if newstmt:
            return (newstmt, [row[n] for n in names])
        names = []
        newstmt = stmt
        while True:
            start = newstmt.find('%(')
            if start == -1:
                break
            end = newstmt.find(')s', start)
            name = newstmt[start+2 : end]
            names.append(name)
            newstmt = newstmt.replace(newstmt[start:end+2], '?',1)#Replace once!
        self.__translations[stmt] = (newstmt, names)
        return (newstmt, [row[n] for n in names])

    def _translate2numeric(self, stmt, row=None):
        # Translate %(name)s to 1,2,... and build a list of attributes to
        # extract from row. Cache both.
        (newstmt, names) = self.__translations.get(stmt, (None, None))
        if newstmt:
            return (newstmt, [row[n] for n in names])
        names = []
        cnt = 0
        newstmt = stmt
        while True:
            start = newstmt.find('%(')
            if start == -1:
                break
            end = newstmt.find(')s', start)
            name = newstmt[start+2 : end]
            names.append(name)
            newstmt = newstmt.replace(newstmt[start:end+2], ':' + str(cnt))
            cnt += 1
        self.__translations[stmt] = (newstmt, names)
        return (newstmt, [row[n] for n in names])
        
    def _translate2format(self, stmt, row=None):
        # Translate %(name)s to %s and build a list of attributes to extract
        # from row. Cache both.
        (newstmt, names) = self.__translations.get(stmt, (None, None))
        if newstmt:
            return (newstmt, [row[n] for n in names])
        names = []
        newstmt = stmt
        while True:
            start = newstmt.find('%(')
            if start == -1:
                break
            end = newstmt.find(')s', start)
            name = newstmt[start+2 : end]
            names.append(name)
            newstmt = newstmt.replace(newstmt[start:end+2],'%s',1)#Replace once!
        self.__translations[stmt] = (newstmt, names)
        return (newstmt, [row[n] for n in names])

    def rowfactory(self, names=None):
        self.__queue.join()
        rows = self.__cursor
        self.__cursor = self.__connection.cursor()
        if names is None:
            if rows.description is None: # no query was executed ...
                return (nothing for nothing in []) # a generator with no rows
            else:
                names = [self.nametranslator(t[0]) for t in rows.description]
        return rowfactory(rows, names, True)

    def fetchone(self, names=None):
        self.__queue.join()
        if self.__cursor.description is None:
            return {}
        if names is None:
            names = [self.nametranslator(t[0]) \
                         for t in self.__cursor.description]
        values = self.__cursor.fetchone()
        if values is None:
            return dict([(n, None) for n in names])#A row with each att = None
        else:
            return dict(zip(names, values))

    def fetchonetuple(self):
        self.__queue.join()
        if self.__cursor.description is None:
            return ()
        values = self.__cursor.fetchone()
        if values is None:
            return (None, ) * len(self.__cursor.description)
        else:
            return values

    def fetchmanytuples(self, cnt):
        self.__queue.join()
        if self.__cursor.description is None:
            return []
        return self.__cursor.fetchmany(cnt)

    def fetchalltuples(self):
        self.__queue.join()
        if self.__cursor.description is None:
            return []
        return self.__cursor.fetchall()

    def rowcount(self):
        self.__queue.join()
        return self.__cursor.rowcount

    def getunderlyingmodule(self):
        # No need to join the queue here
        return modules[self.__connection.__class__.__module__]

    def commit(self):
        endload()
        self.__queue.join()
        self.__connection.commit()

    def close(self):
        self.__queue.join()
        self.__connection.close()

    def rollback(self):
        self.__queue.join()
        self.__connection.rollback()

    def setasdefault(self):
        global _defaulttargetconnection
        _defaulttargetconnection = self

    def cursor(self):
        self.__queue.join()
        return self.__connection.cursor()

    def resultnames(self):
        self.__queue.join()
        if self.__cursor.description is None:
            return None
        else:
            return tuple([t[0] for t in self.__cursor.description])

    def __getstate__(self):
        # In case the ConnectionWrapper is pickled (to be sent to another
        # process), we need to create a new cursor when it is unpickled.
        res = self.__dict__.copy()
        del res['_ConnectionWrapper__cursor'] # a dirty trick, but...

    def __setstate__(self, dict):
        self.__dict__.update(dict)
        self.__cursor = self.__connection.cursor()

    def __worker(self):
        while True:
            (op, curs, stmt, args) = self.__queue.get()
            if op == self._SINGLE:
                curs.execute(stmt, args)
            elif op == self._MANY:
                curs.executemany(stmt, args)
            self.__queue.task_done()


class Error(exceptions.StandardError):
    pass


class InterfaceError(Error):
    pass